/*
    This file is part of libqtelegram library.

    Telegram-client is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    Telegram-client is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this telegram-client.  If not, see <http://www.gnu.org/licenses/>.

    Copyright Roberto Mier & Tiago Herrmann & Vitaly Valtman 2014
*/
#include "connection.h"

#include <sys/socket.h>
#include <netinet/tcp.h>
#include <netinet/in.h>

#include <QDebug>

Connection::Connection(const QString &host, qint32 port, QObject *parent) :
    QTcpSocket(parent),
    Endpoint(host, port) {
}

Connection::Connection(QObject *parent) :
    QTcpSocket(parent) {
}

Connection::~Connection() {
}

qint64 Connection::writeOut(const void *data, qint64 length){
    if (!length) { return 0; }
    Q_ASSERT(length > 0);
    return write((const char *)data, length);
}

qint32 Connection::peekIn(void *data, qint32 len) {
    if (!len || bytesAvailable() < 1) { return 0; }
    Q_ASSERT(len > 0);
    if (len > bytesAvailable()) {
        len = bytesAvailable();
    }
    return peek((char *)data, len);
}

qint32 Connection::readIn(void *data, qint32 len) {
    if (!len || bytesAvailable() < 1) { return 0; }
    Q_ASSERT(len > 0);
    if (len > bytesAvailable()) {
        len = bytesAvailable();
    }
    return read((char *)data, len);
}

QByteArray Connection::readIn(qint32 len) {
    if (!len || bytesAvailable() < 1) { return 0; }
    Q_ASSERT(len > 0);
    if (len > bytesAvailable()) {
        len = bytesAvailable();
    }
    return read(len);
}

QByteArray Connection::readAll() {
    if (bytesAvailable() < 1) { return 0; }
    return readAll();
}

void Connection::connectToServer() {
    Q_ASSERT(!m_host.isEmpty());
    Q_ASSERT(m_port);
    connect(this, SIGNAL(connected()), this, SLOT(onConnected()));
    connect(this, SIGNAL(readyRead()), this, SLOT(onReadyRead()));
    connectToHost(m_host, m_port);
}

void Connection::onConnected() {
    // abridged version of the protocol requires sending 0xef byte at beginning
    char byte = 0xef;
    qint32 writtenBytes = writeOut(&byte, 1);
    Q_UNUSED(writtenBytes);
    Q_ASSERT(writtenBytes == 1);
    // process the rest of operations in inherited classes
    processConnected();
}

void Connection::onReadyRead() {
    qint64 availableBytes = bytesAvailable();
    while (availableBytes > 0) {
        unsigned len = 0;
        unsigned t = 0;
        qint32 readed = peekIn(&len, 1);
        Q_UNUSED(readed);
        Q_ASSERT(readed == 1);
        if (len >= 1 && len <= 0x7e) {
            if (availableBytes < (qint32)(1 + 4 * len)) { return; }
        } else {
            if (availableBytes < 4) { return; }
            readed = peekIn(&len, 4);
            Q_ASSERT(readed == 4);
            len = (len >> 8);
            if (availableBytes < (qint32)(4 + 4 * len)) { return; }
            len = 0x7f;
        }

        if (len >= 1 && len <= 0x7e) {
            readed = readIn(&t, 1);
            Q_ASSERT(readed == 1);
            Q_ASSERT(t == len);
            Q_ASSERT(len >= 1);
        } else {
            Q_ASSERT(len == 0x7f);
            readed = readIn(&len, 4);
            Q_ASSERT(readed == 4);
            len = (len >> 8);
            Q_ASSERT(len >= 1);
        }
        // readed length is real length divided by 4
        len *= 4;
        qint32 op;
        readed = peekIn(&op, 4);
        Q_ASSERT(readed == 4);
        // process rpc answer in inherited classes
        processRpcAnswer(op, len);
        availableBytes = bytesAvailable();
    }
}
