/*
    This file is part of libqtelegram library.

    Telegram-client is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    Telegram-client is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this telegram-client.  If not, see <http://www.gnu.org/licenses/>.

    Copyright Roberto Mier & Tiago Herrmann & Vitaly Valtman 2014
*/
#include "dcprovider.h"
#include "constants.h"
#include "session.h"

DcProvider::DcProvider() : m_api(0), m_pendingDcs(0), m_pendingTransferSessions(0) {
}

DcProvider::~DcProvider() {
    Q_FOREACH (DC *dc, m_dcs) {
        delete dc;
        dc = 0;
    }
    Q_FOREACH (DCAuth *dcAuth, m_dcAuths) {
        if (dcAuth) {
            if (dcAuth->state() == QAbstractSocket::ConnectedState) {
                dcAuth->disconnectFromHost();
            }
            dcAuth->deleteLater();
        }
    }
    Q_FOREACH (Session *session, m_transferSessions)  {
        if (session) {
            if (session->state() == QAbstractSocket::ConnectedState) {
                session->disconnectFromHost();
            }
            session->deleteLater();
        }
    }
    if (m_api) {
        delete m_api;
        m_api = 0;
    }
}

DC *DcProvider::getDc(qint32 dcNum) const {
    return m_dcs.value(dcNum, NULL);
}

DC *DcProvider::getWorkingDc() const {
    return m_dcs.value(Settings::getInstance()->workingDcNum(), NULL);
}


void DcProvider::initialize() {
    /**
      1.- Read settings (dcsList and working dc num)
      2.- Fullfill m_dcs map with dcsList
      3.- Connect to working dc
      4.- Ask for config when working dc is connected
      5.- Connect to the rest of the dcs received in config
      */
    // 1.- Get the settings dcs, fullfill m_dcs map and current dc num
    Settings *s = Settings::getInstance();
    QList<DC *> dcsList = s->dcsList();
    Q_FOREACH (DC *dc, dcsList) {
        m_dcs.insert(dc->id(), dc);
    }
    // 2.- connect to working DC
    if (s->workingDcNum() == 1) {
        // If dcId == 1, it's the default one, so call default host and port
        if (!m_dcs.value(1, 0)) {
            m_dcs[1] = new DC(1);
        }
        // if dc hasn't key created, create it
        if (m_dcs[1]->state() < DC::authKeyCreated) {
            // create a dc authenticator based in dc info
            s->testMode() ? m_dcs[1]->setHost(LIBQTELEGRAM_SERVER_TEST) : m_dcs[1]->setHost(LIBQTELEGRAM_SERVER);
            m_dcs[1]->setPort(443);
            DCAuth *dcAuth = new DCAuth(m_dcs[1], this);
            m_dcAuths.insert(1, dcAuth);
            connect(dcAuth, SIGNAL(dcReady(DC*)), this, SLOT(onDcReady(DC*)));
            dcAuth->createAuthKey();
        } else {
            onDcReady(m_dcs[1]);
        }

    } else {
        // In any other case, the host and port have been retrieved from auth file settings and the DC object is already created
        if (m_dcs[s->workingDcNum()]->state() < DC::authKeyCreated) {
            // create a dc authenticator based in dc info
            DCAuth *dcAuth = new DCAuth(m_dcs[s->workingDcNum()]);
            m_dcAuths.insert(s->workingDcNum(), dcAuth);
            connect(dcAuth, SIGNAL(dcReady(DC*)), this, SLOT(onDcReady(DC*)));
            dcAuth->createAuthKey();
        } else {
            onDcReady(m_dcs[s->workingDcNum()]);
        }
    }
}

void DcProvider::onDcReady(DC *dc) {
    qDebug() << "DC" << dc->id() << "ready";
    // disconnect dcAuth to avoid consuming bandwidth with unnecessary connections
    DCAuth *dcAuth = m_dcAuths.value(dc->id());
    if (dcAuth) {
        if (dcAuth->state() != QAbstractSocket::UnconnectedState) {
            connect(dcAuth, SIGNAL(disconnected()), this, SLOT(onDcAuthDisconnected()));
            dcAuth->disconnectFromHost();
        } else {
            processDcReady(dc);
        }
    } else {
        processDcReady(dc);
    }
}

void DcProvider::onDcAuthDisconnected() {
    DCAuth *dcAuth = qobject_cast<DCAuth *>(sender());
    processDcReady(dcAuth->dc());
    dcAuth->deleteLater();
}

void DcProvider::processDcReady(DC *dc) {
    // create api object if dc is workingDc, and get configuration
    if ((!m_api) && (dc->id() == Settings::getInstance()->workingDcNum())) {
        Session *session = new Session(dc, this);
        m_api = new Api(session, this);
        connect(session, SIGNAL(sessionReady(DC*)), this, SLOT(onApiReady(DC*)));
        session->connectToServer();
    } else if (--m_pendingDcs == 0) { // if all dcs are authorized, emit provider ready signal
        // save the settings here, after all dcs are ready
        Settings::getInstance()->setDcsList(m_dcs.values());
        Settings::getInstance()->saveSettings();

        qDebug() << "DcProvider ready";
        Q_EMIT dcProviderReady();

        // if current dc is in "authKeyCreated" state, authNeeded signal must be emitted for user to sign in.
        // if current dc is in "userSignedIn" state, check that all dcs are in that state or export/import auth.
        // If they aren't, transfer auth from workingDc to workingDc+1, from workingDc+1 to workingDc+2...etc until completed all.
        switch (m_dcs.value(Settings::getInstance()->workingDcNum())->state()) {
        case DC::authKeyCreated:
            Q_EMIT authNeeded();
            break;
        case DC::userSignedIn:
            transferAuth();
            break;
        default:
            Q_ASSERT(0); // it's impossible getting here. This is done to avoid warning when compilation
            break;
        }

    }
}

void DcProvider::onApiReady(DC*) {
    // get the config
    connect(m_api, SIGNAL(config(qint64,qint32,bool,qint32,QList<DcOption>,qint32,qint32)), this, SLOT(onConfigReceived(qint64,qint32,bool,qint32,QList<DcOption>,qint32,qint32)));
    m_api->helpGetConfig();
}

void DcProvider::onConfigReceived(qint64 msgId, qint32 date, bool testMode, qint32 thisDc, const QList<DcOption> &dcOptions, qint32 chatMaxSize, qint32 broadcastMaxSize) {
    qDebug() << "onConfigReceived(), msgId =" << QString::number(msgId, 16);
    qDebug() << "date =" << date;
    qDebug() << "testMode =" << testMode;
    qDebug() << "thisDc =" << thisDc;
    m_pendingDcs = dcOptions.length() -1; //all the received options but the default one, yet used
    Q_FOREACH (DcOption dcOption, dcOptions) {
        qDebug() << "dcOption | id =" << dcOption.id() << ", ipAddress =" << dcOption.ipAddress() <<
                    ", port =" << dcOption.port() << ", hostname =" << dcOption.hostname();

        // for every new DC or not authenticated DC, insert into m_dcs and authenticate
        DC *dc = m_dcs.value(dcOption.id());
        // check if dc is not null or if received host and port are not equals than settings ones
        if ((!dc) || ((dc->host() != dcOption.ipAddress()) || (dc->port() != dcOption.port()))) {
            // if not exists dc or host and port different, create a new dc object for this dcId and add it to m_dcs map
            dc = new DC(dcOption.id());
            dc->setHost(dcOption.ipAddress());
            dc->setPort(dcOption.port());
            m_dcs.insert(dcOption.id(), dc);
        }

        // let's see if needed to create shared key for it
        // In any other case, the host and port have been retrieved from auth file settings and the DC object is already created
        if (dc->state() < DC::authKeyCreated) {
            // create a dc authenticator based in dc info
            DCAuth *dcAuth = new DCAuth(dc, this);
            m_dcAuths.insert(dcOption.id(), dcAuth);
            connect(dcAuth, SIGNAL(dcReady(DC*)), this, SLOT(onDcReady(DC*)));
            dcAuth->createAuthKey();
        } else if (dcOption.id() != thisDc) {
            // if authorized and not working dc emit dcReady signal directly
            onDcReady(dc);
        }
    }
    qDebug() << "chatMaxSize =" << chatMaxSize;
    qDebug() << "broadcastMaxSize =" << broadcastMaxSize;
}

Api *DcProvider::getApi() const {
    return m_api;
}

QList<DC *> DcProvider::getDcs() const {
    return m_dcs.values();
}

void DcProvider::transferAuth() {
    // add in a list all the dcs in DcState::authKeyCreated but not in DCState::userSignedIn
    // and receiving the authorization data (all but working dc one)
    Q_ASSERT(m_api);
    bool hasTransferSessions = false;
    m_workingDcSession = m_api->mainSession();
    Q_ASSERT(m_workingDcSession);
    QList<DC *> m_dcsList = m_dcs.values();
    for (qint32 i=0 ; i < m_dcsList.size(); i++) {
        DC *dc = m_dcsList.value(i);
        if (dc->state() == DC::authKeyCreated &&
                dc->id() != Settings::getInstance()->workingDcNum()) {
            hasTransferSessions = true;
            // create a new session for this dc
            Session *session = m_api->createSession(dc);
            connect(session, SIGNAL(sessionReady(DC*)), this, SLOT(onTransferSessionReady(DC*)));
            session->connectToServer();
            m_pendingTransferSessions++;
        }
    }
    if (!hasTransferSessions) {
        Q_EMIT authTransferCompleted();
    }
}

void DcProvider::onTransferSessionReady(DC *) {
    Session *session = qobject_cast<Session *>(sender());
    m_transferSessions.append(session);
    if (--m_pendingTransferSessions == 0) {
        connect(m_api, SIGNAL(authExportedAuthorization(qint64,qint32,QByteArray)), this, SLOT(onAuthExportedAuthorization(qint64,qint32,QByteArray)));
        connect(m_api, SIGNAL(authImportedAuthorization(qint64,qint32,User)), this, SLOT(onAuthImportedAuthorization(qint64,qint32,User)));
        m_api->authExportAuthorization(m_transferSessions.first()->dc()->id());
    }
}

void DcProvider::onAuthExportedAuthorization(qint64, qint32 ourId, const QByteArray &bytes) {
    // Set ourId into settings (It doesn't matter if set before)
    Settings::getInstance()->setOurId(ourId);
    // Change api dc to first in the transfer dcs list
    m_api->setMainSession(m_transferSessions.first());
    // Execute import in this dc
    m_api->authImportAuthorization(ourId, bytes);
}

void DcProvider::onAuthImportedAuthorization(qint64, qint32 expires, const User &) {
    Session *s = m_transferSessions.takeFirst();
    DC *authorizedDc = s->dc();
    authorizedDc->setExpires(expires);
    authorizedDc->setState(DC::userSignedIn);
    s->disconnectFromHost();
    s->deleteLater();
    m_api->setMainSession(m_workingDcSession);
    if (m_transferSessions.isEmpty()) {
        Settings::getInstance()->setDcsList(m_dcs.values());
        Settings::getInstance()->saveSettings();
        Q_EMIT authTransferCompleted();
    } else {
        m_api->authExportAuthorization(m_transferSessions.first()->dc()->id());
    }
}
