#ifndef SESSION_H
#define SESSION_H

#include "dc.h"
#include "connection.h"
#include "query.h"
#include <QMap>
#include <QLoggingCategory>

Q_DECLARE_LOGGING_CATEGORY(TG_NET_SESSION)

class Session : public Connection
{
    Q_OBJECT
public:
    explicit Session(DC *dc, QObject *parent = 0);
    ~Session();

    DC *dc();
    qint64 sendQuery(OutboundPkt &outboundPkt, QueryMethods *methods, void *extra = 0);

    inline qint64 sessionId() { return m_sessionId; }
    inline bool initConnectionNeeded() { return m_initConnectionNeeded; }
    inline void setInitConnectionNeeded(bool initConnectionNeeded) { m_initConnectionNeeded = initConnectionNeeded; }
    inline qint64 clientLastMsgId() { return m_clientLastMsgId; }
    inline qint32 seqNo() { return m_seqNo; }
    inline void setSeqNo(qint32 seqNo) { m_seqNo = seqNo; }

Q_SIGNALS:
    void sessionReady(DC *dc);

    void resultReceived(Query *q, InboundPkt &inboundPkt);
    void errorReceived(Query *q, qint32 errorCode, QString errorText);

    void updatesTooLong();
    void updateShortMessage(qint32 id, qint32 fromId, const QString &message, qint32 pts, qint32 date, qint32 seq);
    void updateShortChatMessage(qint32 id, qint32 fromId, qint32 chatId, const QString &message, qint32 pts, qint32 date, qint32 seq);
    void updateShort(Update update, qint32 date);
    void updatesCombined(QList<Update> updates, QList<User> users, QList<Chat> chats, qint32 date, qint32 seqStart, qint32 seq);
    void updates(QList<Update> udts, QList<User> users, QList<Chat> chats, qint32 date, qint32 seq);

private:
    struct EncryptedMsg {
        // unencrypted header
        qint64 authKeyId;
        char msgKey[16];
        // encrypted part, starts with encrypted header
        qint64 serverSalt;
        qint64 sessionId;
        // first message follows
        qint64 msgId;
        qint32 seqNo;
        qint32 msgLen;   // divisible by 4
        qint32 message[MAX_MESSAGE_INTS];
    };


    // session members
    qint64 m_sessionId;
    qint64 m_serverSalt;
    qint32 m_timeSyncFactor; // difference between client and server time
    static qint64 m_clientLastMsgId;
    qint32 m_seqNo;

    DC *m_dc;
    QMap<qint64, struct Query *> m_pendingQueries; // map with the msgIds of the queries sent but not received its ack yet.

    // Additional flags:
    // flag to see if initConnection() method must be sent before next output rpc query
    // (see https://core.telegram.org/api/invoking#saving-client-info)
    // This is needed at first operation against a new dc (and at beginning)
    // (the flag is set to true when just connected to dc until initConnection() operation is executed against this server)
    bool m_initConnectionNeeded;

    // connected signal management
    void processConnected();

    // process and transform response message
    void processRpcAnswer(qint32 op, qint32 len);
    void processRpcMessage(InboundPkt &inboundPkt);

    // execute rpc answer operation requested
    void rpcExecuteAnswer(InboundPkt &inboundPkt, qint64 msgId);

    void workUpdate(InboundPkt &inboundPkt, qint64 msgId);
    void workContainer (InboundPkt &inboundPkt, qint64 msgId);
    void workNewSessionCreated(InboundPkt &inboundPkt, qint64 msgId);
    void workMsgsAck(InboundPkt &inboundPkt, qint64 msgId);
    void workRpcResult(InboundPkt &inboundPkt, qint64 msgId);
    void workUpdateShort(InboundPkt &inboundPkt, qint64 msgId);
    void workUpdates(InboundPkt &inboundPkt, qint64 msgId);
    void workUpdatesCombined(InboundPkt &inboundPkt, qint64 msgId);
    void workUpdateShortMessage(InboundPkt &inboundPkt, qint64 msgId);
    void workUpdateShortChatMessage(InboundPkt &inboundPkt, qint64 msgId);
    void workPacked(InboundPkt &inboundPkt, qint64 msgId);
    void workBadServerSalt(InboundPkt &inboundPkt, qint64 msgId);
    void workPong(InboundPkt &inboundPkt, qint64 msgId);
    void workDetailedInfo(InboundPkt &inboundPkt, qint64 msgId);
    void workNewDetailedInfo(InboundPkt &inboundPkt, qint64 msgId);
    void workUpdatesTooLong(InboundPkt &inboundPkt, qint64 msgId);
    void workBadMsgNotification(InboundPkt &inboundPkt, qint64 msgId);

    // encrypt message and send
    EncryptedMsg *initEncryptedMessage(qint32 useful);
    qint64 encryptSendMessage(qint32 *msg, qint32 msgInts, qint32 useful);
    qint32 aesEncryptMessage(EncryptedMsg *encMsg);
    void rpcSendMessage(void *data, qint32 len);
    qint64 generateNextMsgId();

    // query results
    void queryOnResult(InboundPkt &inboundPkt, qint64 msgId);
    void queryOnError(InboundPkt &inboundPkt, qint64 msgId);

    void recomposeAndSendQuery(Query *q);

private Q_SLOTS:
    void resendQuery(Query *q);

};

#endif // SESSION_H
