/*
    This file is part of libqtelegram library.

    Telegram-client is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    Telegram-client is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this telegram-client.  If not, see <http://www.gnu.org/licenses/>.

    Copyright Roberto Mier & Tiago Herrmann & Vitaly Valtman 2014
*/
#include "settings.h"
#include "utils.h"
#include "constants.h"
#include "dc.h"

#include <QDir>
#include <QSettings>
#include <QDebug>
#include <fcntl.h>
#include <unistd.h>

Settings::Settings() {
}

Settings::~Settings() {
    if (m_pubKey) {
        delete m_pubKey;
    }
}

Settings* Settings::getInstance() {
    static Settings *instance = new Settings;
    return instance;
}

bool Settings::loadSettings(const QString &phoneNumber, const QString &baseConfigDirectory, const QString &publicKeyFile) {
    m_phoneNumber = phoneNumber;
    QString configDirectory = baseConfigDirectory + "/" + m_phoneNumber;
    qDebug() << "loading configuration in path:" << configDirectory << "taking as public key file:" << publicKeyFile;
    m_pubKey = Utils::rsaLoadPublicKey(publicKeyFile);
    m_pkFingerprint = Utils::computeRSAFingerprint(m_pubKey);

    // read and normalize config path
    QString configPath = QString(configDirectory).replace("~",QDir::homePath());
    QString configFilename = configPath + '/' + CONFIG_FILE;
    QDir configDir = QDir(configPath);
    if (!configDir.exists()) {
        if (QDir::root().mkpath(configPath)) {
            qDebug() << configPath << "config directory created";
        } else {
            qWarning() << "error creating config directory: " << configPath;
            return false;
        }
    }

    //check if config file is there and it's readable
    QFile configFile(configFilename);
    if (!configFile.exists()) {
        // config file missing, touch it
        if (configFile.open(QIODevice::WriteOnly)) {
            QTextStream out(&configFile);
            out << DEFAULT_CONFIG_CONTENTS;
            configFile.close();
        } else {
            qWarning() << "Impossible to open file " << configFilename << " for writting default configuration";
            return false;
        }
    }

    QString downloadsDirname = configPath + '/' + DOWNLOADS_DIRECTORY;
    QDir downloadsDir = QDir(downloadsDirname);
    if (!downloadsDir.exists()) {
        if (QDir::root().mkpath(downloadsDirname)) {
            qDebug() << downloadsDirname << "downloads directory created";
        } else {
            qWarning() << "error creating downloads directory: " << downloadsDirname;
            return false;
        }
    }

    //read settings
    qDebug() << "loading settings file" << configFilename;
    QSettings settings(configFilename, QSettings::IniFormat);
    m_testMode = settings.value(ST_TEST_MODE, false).toBool();
    m_managedDownloads = settings.value(ST_MANAGED_DOWNLOADS, false).toBool();
    m_langCode = settings.value(ST_LANG_CODE, "en").toString();
    m_authFilename = settings.value(ST_AUTH_FILE, configPath + '/' + AUTH_KEY_FILE).toString();
    m_secretChatFilename = settings.value(ST_SECRET, configPath + '/' + SECRET_CHAT_FILE).toString();
    m_downloadsDirectory = settings.value(ST_DOWNLOADS, configPath + '/' + DOWNLOADS_DIRECTORY).toString();
    m_stateFilename = settings.value(ST_STATE_FILE, configPath + '/' + STATE_FILE).toString();

    // log readed data
    qDebug() << "testMode:" << m_testMode;
    qDebug() << "managedDownloads:" << m_managedDownloads;
    qDebug() << "lang:" << m_langCode;
    qDebug() << "authFile:" << m_authFilename;
    qDebug() << "secretChatFile:" << m_secretChatFilename;
    qDebug() << "downloads:" << m_downloadsDirectory;
    qDebug() << "stateFile:" << m_stateFilename;

    readAuthFile();

    return true;
}

void Settings::saveSettings() {
    writeAuthFile();
}

void Settings::writeAuthFile() {
    QSettings settings(m_authFilename, QSettings::IniFormat);
    Settings::getInstance()->testMode() ? settings.beginGroup(ST_TEST) : settings.beginGroup(ST_PRODUCTION);
    settings.setValue(ST_WORKING_DC_NUM, m_workingDcNum);
    settings.setValue(ST_OUR_ID, m_ourId);
    settings.beginWriteArray(ST_DCS_ARRAY);
    for (qint32 i = 0; i < m_dcsList.length(); i++) {
        //write dc
        settings.setArrayIndex(i);
        settings.setValue(ST_DC_NUM, m_dcsList[i]->id());
        settings.setValue(ST_HOST, m_dcsList[i]->host());
        settings.setValue(ST_PORT, m_dcsList[i]->port());
        settings.setValue(ST_DC_STATE, m_dcsList[i]->state());
        if (m_dcsList[i]->authKeyId()) {
            settings.setValue(ST_AUTH_KEY_ID, m_dcsList[i]->authKeyId());
            QByteArray baToSave(m_dcsList[i]->authKey(), 256);
            settings.setValue(ST_AUTH_KEY, baToSave.toBase64());
        }
        settings.setValue(ST_SERVER_SALT, m_dcsList[i]->serverSalt());
        settings.setValue(ST_EXPIRES, m_dcsList[i]->expires());
    }
    settings.endArray();
    settings.endGroup();
}

void Settings::readAuthFile() {
    QSettings settings(m_authFilename, QSettings::IniFormat);
    Settings::getInstance()->testMode() ? settings.beginGroup(ST_TEST) : settings.beginGroup(ST_PRODUCTION);
    m_workingDcNum = settings.value(ST_WORKING_DC_NUM, 1).toInt();
    m_ourId = settings.value(ST_OUR_ID).toInt();

    qDebug() << "workingDcNum:" << m_workingDcNum;
    qDebug() << "ourId:" << m_ourId;

    // read all dcs
    m_dcsList.clear();
    qint32 n = settings.beginReadArray(ST_DCS_ARRAY);
    for (qint32 i = 0; i < n; i++) {
        settings.setArrayIndex(i);
        qint32 dcNum = settings.value(ST_DC_NUM).toInt();
        DC* dc = new DC(dcNum);
        dc->setHost(settings.value(ST_HOST).toString());
        dc->setPort(settings.value(ST_PORT, 0).toInt());
        dc->setState((DC::DcState)settings.value(ST_DC_STATE, DC::init).toInt());
        dc->setAuthKeyId(settings.value(ST_AUTH_KEY_ID, 0).toLongLong());
        if (dc->state() >= DC::authKeyCreated) {
            QByteArray readedBa = QByteArray::fromBase64(settings.value(ST_AUTH_KEY).toByteArray());
            memcpy(dc->authKey(), readedBa.data(), 256);
        }
        dc->setServerSalt(settings.value(ST_SERVER_SALT, 0).toLongLong());
        dc->setExpires(settings.value(ST_EXPIRES).toInt());

        qDebug() << "DC | id:" << dc->id() << ", state:" << dc->state() <<
                    ", host:" << dc->host() << ", port:" << dc->port() <<
                    ", expires:" << dc->expires() << ", authKeyId:" << dc->authKeyId() <<
                    ", serverSalt:" << dc->serverSalt();

        m_dcsList.insert(dcNum, dc);
    }
    settings.endArray();
    settings.endGroup();
}
