/*
    This file is part of libqtelegram library.

    Telegram-client is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    Telegram-client is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this telegram-client.  If not, see <http://www.gnu.org/licenses/>.

    Copyright Roberto Mier & Tiago Herrmann & Vitaly Valtman 2014
*/

#include "core/dcprovider.h"
#include "telegram.h"

#include <QDebug>
#include <QLoggingCategory>
#include <QCryptographicHash>
#include <QFile>
#include <QFileInfo>
#include <QMimeDatabase>

Telegram::Telegram(const QString phoneNumber, const QString configPath, const QString publicKeyFile) {
    // Qt5.2 doesn't support .ini files to control logging, so use this
    // manual workaround instead.
    // http://blog.qt.digia.com/blog/2014/03/11/qt-weekly-1-categorized-logging/
    QLoggingCategory::setFilterRules(QString(qgetenv("QT_LOGGING_RULES")));

    // load settings
    Settings::getInstance()->loadSettings(phoneNumber, configPath, publicKeyFile);
    // activate dc provider ready signal
    connect(&m_dcProvider, SIGNAL(dcProviderReady()), this, SLOT(onDcProviderReady()));
    m_dcProvider.initialize();
    // activate rest of dc provider signal connections
    connect(&m_dcProvider, SIGNAL(authNeeded()), this, SIGNAL(authNeeded()));
    connect(&m_dcProvider, SIGNAL(authTransferCompleted()), this, SIGNAL(authLoggedIn()));
}

Telegram::~Telegram() {
    // delete any remainig download files, upload files or operations
    Q_FOREACH (UploadFile *f, m_filesMap) {
        if (f) delete f;
    }
    Q_FOREACH (const FileOperation *op, m_fileOperationsMap) {
        if (op) delete op;
    }
    Q_FOREACH (const DownloadFile *f, m_downloadsMap) {
        if (f) delete f;
    }
}

qint32 Telegram::ourId() {
    return Settings::getInstance()->ourId();
}

void Telegram::onDcProviderReady() {
    m_api = m_dcProvider.getApi();
    // api signal-signal and signal-slot connections
    // updates
    connect(m_api, SIGNAL(updatesTooLong()), this, SIGNAL(updatesTooLong()));
    connect(m_api, SIGNAL(updateShortMessage(qint32,qint32,QString,qint32,qint32,qint32)), this, SIGNAL(updateShortMessage(qint32,qint32,QString,qint32,qint32,qint32)));
    connect(m_api, SIGNAL(updateShortChatMessage(qint32,qint32,qint32,QString,qint32,qint32,qint32)), this, SIGNAL(updateShortChatMessage(qint32,qint32,qint32,QString,qint32,qint32,qint32)));
    connect(m_api, SIGNAL(updateShort(Update,qint32)), this, SIGNAL(updateShort(Update,qint32)));
    connect(m_api, SIGNAL(updatesCombined(QList<Update>,QList<User>,QList<Chat>,qint32,qint32,qint32)), this, SIGNAL(updatesCombined(QList<Update>,QList<User>,QList<Chat>,qint32,qint32,qint32)));
    connect(m_api, SIGNAL(updates(QList<Update>,QList<User>,QList<Chat>,qint32,qint32)), this, SIGNAL(updates(QList<Update>,QList<User>,QList<Chat>,qint32,qint32)));
    // errors
    connect(m_api, SIGNAL(error(qint64,qint32,QString)), this, SIGNAL(error(qint64,qint32,QString)));
    connect(m_api, SIGNAL(authCheckPhoneError(qint64,qint32,QString)), this, SLOT(onAuthCheckPhoneError(qint64,qint32,QString)));
    connect(m_api, SIGNAL(authSignInError(qint64,qint32,QString)), this, SIGNAL(authSignInError(qint64,qint32,QString)));
    connect(m_api, SIGNAL(authSignUpError(qint64,qint32,QString)), this, SIGNAL(authSignUpError(qint64,qint32,QString)));
    connect(m_api, SIGNAL(uploadFileError(qint64,qint32,QString)), this, SLOT(onUploadGetFileError(qint64,qint32,QString)));
    // positive responses
    connect(m_api, SIGNAL(authCheckedPhone(qint64,bool,bool)), this, SIGNAL(authCheckPhoneAnswer(qint64,bool,bool)));
    connect(m_api, SIGNAL(authSentCode(qint64,bool,QString,qint32,bool)), this, SLOT(onAuthSentCode(qint64,bool,QString,qint32,bool)));
    connect(m_api, SIGNAL(authSendCallResult(qint64,bool)), this, SIGNAL(authSendCallAnswer(qint64,bool)));
    connect(m_api, SIGNAL(authSignInAuthorization(qint64,qint32,User)), this, SLOT(onUserAuthorized(qint64,qint32,User)));
    connect(m_api, SIGNAL(authSignUpAuthorization(qint64,qint32,User)), this, SLOT(onUserAuthorized(qint64,qint32,User)));
    connect(m_api, SIGNAL(authLogOutResult(qint64,bool)), this, SIGNAL(authLogOutAnswer(qint64,bool)));
    connect(m_api, SIGNAL(authSendInvitesResult(qint64,bool)), this, SIGNAL(authSendInvitesAnswer(qint64,bool)));
    connect(m_api, SIGNAL(authResetAuthorizationsResult(qint64,bool)), this, SIGNAL(authResetAuthorizationsAnswer(qint64,bool)));
    connect(m_api, SIGNAL(accountRegisterDeviceResult(qint64,bool)), this, SIGNAL(accountRegisterDeviceAnswer(qint64,bool)));
    connect(m_api, SIGNAL(accountUnregisterDeviceResult(qint64,bool)), this, SIGNAL(accountUnregisterDeviceAnswer(qint64,bool)));
    connect(m_api, SIGNAL(accountUpdateNotifySettingsResult(qint64,bool)), this, SIGNAL(accountUpdateNotifySettingsAnswer(qint64,bool)));
    connect(m_api, SIGNAL(accountPeerNotifySettings(qint64,PeerNotifySettings)), this, SIGNAL(accountGetNotifySettingsAnswer(qint64,PeerNotifySettings)));
    connect(m_api, SIGNAL(accountResetNotifySettingsResult(qint64,bool)), this, SIGNAL(accountResetNotifySettingsAnswer(qint64,bool)));
    connect(m_api, SIGNAL(accountUser(qint64,User)), this, SIGNAL(accountUpdateProfileAnswer(qint64,User)));
    connect(m_api, SIGNAL(accountUpdateStatusResult(qint64,bool)), this, SIGNAL(accountUpdateStatusAnswer(qint64,bool)));
    connect(m_api, SIGNAL(accountGetWallPapersResult(qint64,QList<WallPaper>)), this, SIGNAL(accountGetWallPapersAnswer(qint64,QList<WallPaper>)));
    connect(m_api, SIGNAL(photosPhoto(qint64,Photo,QList<User>)), this, SIGNAL(photosUploadProfilePhotoAnswer(qint64,Photo,QList<User>)));
    connect(m_api, SIGNAL(photosUserProfilePhoto(qint64,UserProfilePhoto)), this, SIGNAL(photosUpdateProfilePhotoAnswer(qint64,UserProfilePhoto)));
    connect(m_api, SIGNAL(usersGetUsersResult(qint64,QList<User>)), this, SIGNAL(usersGetUsersAnswer(qint64,QList<User>)));
    connect(m_api, SIGNAL(userFull(qint64,User,ContactsLink,Photo,PeerNotifySettings,bool,QString,QString)), this, SIGNAL(usersGetFullUserAnswer(qint64,User,ContactsLink,Photo,PeerNotifySettings,bool,QString,QString)));
    connect(m_api, SIGNAL(photosPhotos(qint64,QList<Photo>,QList<User>)), this, SLOT(onPhotosPhotos(qint64, QList<Photo>, QList<User>)));
    connect(m_api, SIGNAL(photosPhotosSlice(qint64,qint32,QList<Photo>,QList<User>)), this, SIGNAL(photosGetUserPhotosAnswer(qint64,qint32,QList<Photo>,QList<User>)));
    connect(m_api, SIGNAL(contactsGetStatusesResult(qint64,QList<ContactStatus>)), this, SIGNAL(contactsGetStatusesAnswer(qint64,QList<ContactStatus>)));
    connect(m_api, SIGNAL(contactsContacts(qint64,QList<Contact>,QList<User>)), this, SLOT(onContactsContacts(qint64,QList<Contact>,QList<User>)));
    connect(m_api, SIGNAL(contactsContactsNotModified(qint64)), this, SLOT(onContactsContactsNotModified(qint64)));
    connect(m_api, SIGNAL(contactsImportedContacts(qint64,QList<ImportedContact>,QList<qint64>,QList<User>)), this, SIGNAL(contactsImportContactsAnswer(qint64,QList<ImportedContact>,QList<qint64>,QList<User>)));
    connect(m_api, SIGNAL(contactsDeleteContactLink(qint64,ContactsMyLink,ContactsForeignLink,User)), this, SIGNAL(contactsDeleteContactAnswer(qint64,ContactsMyLink,ContactsForeignLink,User)));
    connect(m_api, SIGNAL(contactsDeleteContactsResult(qint64,bool)), this, SIGNAL(contactsDeleteContactsAnswer(qint64,bool)));
    connect(m_api, SIGNAL(contactsBlockResult(qint64,bool)), this, SIGNAL(contactsBlockAnswer(qint64,bool)));
    connect(m_api, SIGNAL(contactsUnblockResult(qint64,bool)), this, SIGNAL(contactsUnblockAnswer(qint64,bool)));
    connect(m_api, SIGNAL(contactsBlocked(qint64,QList<ContactBlocked>,QList<User>)), this, SLOT(onContactsBlocked(qint64,QList<ContactBlocked>,QList<User>)));
    connect(m_api, SIGNAL(contactsBlockedSlice(qint64,qint32,QList<ContactBlocked>,QList<User>)), this, SIGNAL(contactsGetBlockedAnswer(qint64,qint32,QList<ContactBlocked>,QList<User>)));
    connect(m_api, SIGNAL(messagesSentMessage(qint64,qint32,qint32,qint32,qint32)), this, SLOT(onMessagesSentMessage(qint64,qint32,qint32,qint32,qint32)));
    connect(m_api, SIGNAL(messagesSentMessageLink(qint64,qint32,qint32,qint32,qint32,QList<ContactsLink>)), this, SIGNAL(messagesSendMessageAnswer(qint64,qint32,qint32,qint32,qint32,QList<ContactsLink>)));
    connect(m_api, SIGNAL(messagesSentMediaStatedMessage(qint64,Message,QList<Chat>,QList<User>,qint32,qint32)), this, SLOT(onMessagesSendMediaStatedMessage(qint64,Message,QList<Chat>,QList<User>,qint32,qint32)));
    connect(m_api, SIGNAL(messagesSentMediaStatedMessageLink(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)), this, SLOT(onMessagesSendMediaStatedMessageLink(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)));
    connect(m_api, SIGNAL(messagesSetTypingResult(qint64,bool)), this, SIGNAL(messagesSetTypingAnswer(qint64,bool)));
    connect(m_api, SIGNAL(messagesGetMessagesMessages(qint64,QList<Message>,QList<Chat>,QList<User>)), this, SLOT(onMessagesGetMessagesMessages(qint64,QList<Message>,QList<Chat>,QList<User>)));
    connect(m_api, SIGNAL(messagesGetMessagesMessagesSlice(qint64,qint32,QList<Message>,QList<Chat>,QList<User>)), this, SIGNAL(messagesGetMessagesAnswer(qint64,qint32,QList<Message>,QList<Chat>,QList<User>)));
    connect(m_api, SIGNAL(messagesDialogs(qint64, QList<Dialog>,QList<Message>,QList<Chat>,QList<User>)), this, SLOT(onMessagesDialogs(qint64,QList<Dialog>,QList<Message>,QList<Chat>,QList<User>)));
    connect(m_api, SIGNAL(messagesDialogsSlice(qint64,qint32,QList<Dialog>,QList<Message>,QList<Chat>,QList<User>)), this, SIGNAL(messagesGetDialogsAnswer(qint64,qint32,QList<Dialog>,QList<Message>,QList<Chat>,QList<User>)));
    connect(m_api, SIGNAL(messagesGetHistoryMessages(qint64,QList<Message>,QList<Chat>,QList<User>)), this, SLOT(onMessagesGetHistoryMessages(qint64,QList<Message>,QList<Chat>,QList<User>)));
    connect(m_api, SIGNAL(messagesGetHistoryMessagesSlice(qint64,qint32,QList<Message>,QList<Chat>,QList<User>)), this, SIGNAL(messagesGetHistoryAnswer(qint64,qint32,QList<Message>,QList<Chat>,QList<User>)));
    connect(m_api, SIGNAL(messagesSearchMessages(qint64,QList<Message>,QList<Chat>,QList<User>)), this, SLOT(onMessagesSearchMessages(qint64,QList<Message>,QList<Chat>,QList<User>)));
    connect(m_api, SIGNAL(messagesSearchMessagesSlice(qint64,qint32,QList<Message>,QList<Chat>,QList<User>)), this, SIGNAL(messagesSearchAnswer(qint64,qint32,QList<Message>,QList<Chat>,QList<User>)));
    connect(m_api, SIGNAL(messagesReadAffectedHistory(qint64,qint32,qint32,qint32)), this, SIGNAL(messagesReadHistoryAnswer(qint64,qint32,qint32,qint32)));
    connect(m_api, SIGNAL(messagesDeleteAffectedHistory(qint64,qint32,qint32,qint32)), this, SIGNAL(messagesDeleteHistoryAnswer(qint64,qint32,qint32,qint32)));
    connect(m_api, SIGNAL(messagesDeleteMessagesResult(qint64,QList<qint32>)), this, SIGNAL(messagesDeleteMessagesAnswer(qint64,QList<qint32>)));
    connect(m_api, SIGNAL(messagesRestoreMessagesResult(qint64,QList<qint32>)), this, SIGNAL(messagesRestoreMessagesAnswer(qint64,QList<qint32>)));
    connect(m_api, SIGNAL(messagesReceivedMessagesResult(qint64,QList<qint32>)), this, SIGNAL(messagesReceivedMessagesAnswer(qint64,QList<qint32>)));
    connect(m_api, SIGNAL(messagesForwardMsgStatedMessage(qint64,Message,QList<Chat>,QList<User>,qint32,qint32)), this, SLOT(onMessagesForwardMsgStatedMessage(qint64,Message,QList<Chat>,QList<User>,qint32,qint32)));
    connect(m_api, SIGNAL(messagesForwardMsgStatedMessageLink(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)), this, SIGNAL(messagesForwardMessageAnswer(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)));
    connect(m_api, SIGNAL(messagesForwardMsgsStatedMessages(qint64,QList<Message>,QList<Chat>,QList<User>,qint32,qint32)), this, SLOT(onMessagesForwardMsgsStatedMessages(qint64,QList<Message>,QList<Chat>,QList<User>,qint32,qint32)));
    connect(m_api, SIGNAL(messagesForwardMsgsStatedMessagesLinks(qint64,QList<Message>,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)), this, SIGNAL(messagesForwardMessagesAnswer(qint64,QList<Message>,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)));
    connect(m_api, SIGNAL(messagesSendBroadcastStatedMessages(qint64,QList<Message>,QList<Chat>,QList<User>,qint32,qint32)), this, SLOT(onMessagesSendBroadcastStatedMessages(qint64,QList<Message>,QList<Chat>,QList<User>,qint32,qint32)));
    connect(m_api, SIGNAL(messagesSendBroadcastStatedMessagesLinks(qint64,QList<Message>,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)), this, SIGNAL(messagesSendBroadcastAnswer(qint64,QList<Message>,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)));
    connect(m_api, SIGNAL(messagesChats(qint64,QList<Chat>,QList<User>)), this, SIGNAL(messagesGetChatsAnswer(qint64,QList<Chat>,QList<User>)));
    connect(m_api, SIGNAL(messagesChatFull(qint64,ChatFull,QList<Chat>,QList<User>)), this, SIGNAL(messagesGetFullChatAnswer(qint64,ChatFull,QList<Chat>,QList<User>)));
    connect(m_api, SIGNAL(messagesEditChatTitleStatedMessage(qint64,Message,QList<Chat>,QList<User>,qint32,qint32)), this, SLOT(onMessagesEditChatTitleStatedMessage(qint64,Message,QList<Chat>,QList<User>,qint32,qint32)));
    connect(m_api, SIGNAL(messagesEditChatTitleStatedMessageLink(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)), this, SIGNAL(messagesEditChatTitleAnswer(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)));
    connect(m_api, SIGNAL(messagesEditChatPhotoStatedMessage(qint64,Message,QList<Chat>,QList<User>,qint32,qint32)), this, SLOT(onMessagesEditChatPhotoStatedMessage(qint64,Message,QList<Chat>,QList<User>,qint32,qint32)));
    connect(m_api, SIGNAL(messagesEditChatPhotoStatedMessageLink(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)), this, SIGNAL(messagesEditChatPhotoAnswer(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)));
    connect(m_api, SIGNAL(messagesAddChatUserStatedMessage(qint64,Message,QList<Chat>,QList<User>,qint32,qint32)), this, SLOT(onMessagesAddChatUserStatedMessage(qint64,Message,QList<Chat>,QList<User>,qint32,qint32)));
    connect(m_api, SIGNAL(messagesAddChatUserStatedMessageLink(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)), this, SIGNAL(messagesAddChatUserAnswer(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)));
    connect(m_api, SIGNAL(messagesDeleteChatUserStatedMessage(qint64,Message,QList<Chat>,QList<User>,qint32,qint32)), this, SLOT(onMessagesDeleteChatUserStatedMessage(qint64,Message,QList<Chat>,QList<User>,qint32,qint32)));
    connect(m_api, SIGNAL(messagesDeleteChatUserStatedMessageLink(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)), this, SIGNAL(messagesDeleteChatUserAnswer(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)));
    connect(m_api, SIGNAL(messagesCreateChatStatedMessage(qint64,Message,QList<Chat>,QList<User>,qint32,qint32)), this, SLOT(onMessagesCreateChatStatedMessage(qint64,Message,QList<Chat>,QList<User>,qint32,qint32)));
    connect(m_api, SIGNAL(messagesCreateChatStatedMessageLink(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)), this, SIGNAL(messagesCreateChatAnswer(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)));
    connect(m_api, SIGNAL(updatesState(qint64,qint32,qint32,qint32,qint32,qint32)), this, SIGNAL(updatesGetStateAnswer(qint64,qint32,qint32,qint32,qint32,qint32)));
    connect(m_api, SIGNAL(updatesDifferenceEmpty(qint64,qint32,qint32)), this, SIGNAL(updatesGetDifferenceAnswer(qint64,qint32,qint32)));
    connect(m_api, SIGNAL(updatesDifference(qint64,QList<Message>,QList<EncryptedMessage>,QList<Update>,QList<Chat>&,QList<User>&,UpdatesState)), this, SLOT(onUpdatesDifference(qint64,QList<Message>,QList<EncryptedMessage>,QList<Update>,QList<Chat>&,QList<User>&,UpdatesState)));
    connect(m_api, SIGNAL(updatesDifferenceSlice(qint64,QList<Message>,QList<EncryptedMessage>,QList<Update>,QList<Chat>&,QList<User>&,UpdatesState)), this, SLOT(onUpdatesDifferenceSlice(qint64,QList<Message>,QList<EncryptedMessage>,QList<Update>,QList<Chat>&,QList<User>&,UpdatesState)));
    connect(m_api, SIGNAL(uploadSaveFilePartResult(qint64,qint64,bool)), this, SLOT(onUploadSaveFilePartResult(qint64,qint64,bool)));
    connect(m_api, SIGNAL(uploadSaveBigFilePartResult(qint64,qint64,bool)), this, SLOT(onUploadSaveFilePartResult(qint64,qint64,bool)));
    connect(m_api, SIGNAL(uploadFile(qint64,StorageFileType,qint32,QByteArray)), this, SLOT(onUploadGetFileAnswer(qint64,StorageFileType,qint32,QByteArray)));
    // logic additional signal slots
    connect(m_api, SIGNAL(mainSessionDcChanged(DC*)), this, SLOT(onAuthCheckPhoneDcChanged()));
}

// error and internal managements
void Telegram::onAuthCheckPhoneError(qint64 id, qint32 errorCode, const QString &errorText) {
    // check for error and resend authCheckPhone() request
    if (errorText.contains("_MIGRATE_")) {
        qint32 newDc = errorText.mid(errorText.lastIndexOf("_") + 1).toInt();
        qDebug() << "migrated to dc" << newDc;
        Settings::getInstance()->setWorkingDcNum(newDc);
        DC *dc = m_dcProvider.getDc(newDc);
        // changeMainSession() ---> onAuthCheckPhoneDcChanged()
        m_api->changeMainSessionToDc(dc);
    } else {
        Q_EMIT error(id, errorCode, errorText);
    }
}

void Telegram::onAuthCheckPhoneDcChanged() {
    authCheckPhone();
}

void Telegram::onUserAuthorized(qint64, qint32 expires, const User &) {
    // change state of current dc
    qint32 workingDcNum = Settings::getInstance()->workingDcNum();
    DC *dc = m_dcProvider.getDc(workingDcNum);
    dc->setState(DC::userSignedIn);
    dc->setExpires(expires);
    QList<DC *> dcsList = m_dcProvider.getDcs();
    // save the settings here, after user auth ready in current dc
    Settings::getInstance()->setDcsList(dcsList);
    Settings::getInstance()->saveSettings();
    // transfer current dc authorization to other dcs
    m_dcProvider.transferAuth();
}

void Telegram::onPhotosPhotos(qint64 msgId, const QList<Photo> &photos, const QList<User> &users) {
    Q_EMIT photosGetUserPhotosAnswer(msgId, photos.size(), photos, users);
}

void Telegram::onContactsContacts(qint64 msgId, const QList<Contact> &contacts, const QList<User> &users) {
    m_cachedContacts = contacts;
    m_cachedUsers = users;
    Q_EMIT contactsGetContactsAnswer(msgId, true, contacts, users);
}

void Telegram::onContactsContactsNotModified(qint64 msgId) {
    Q_EMIT contactsGetContactsAnswer(msgId, false, m_cachedContacts, m_cachedUsers);
}

void Telegram::onUploadGetFileError(qint64 id, qint32 errorCode, const QString &errorText) {
    // check for error and resend authCheckPhone() request
    if (errorText.contains("_MIGRATE_")) {
        qint32 newDc = errorText.mid(errorText.lastIndexOf("_") + 1).toInt();
        qDebug() << "file migrated to dc" << newDc;
        DC *dc = m_dcProvider.getDc(newDc);
        DownloadFile *f = m_downloadsMap.take(id);
        // disconnect and remove previous session
        f->session()->disconnectFromHost();
        f->session()->deleteLater();
        // create new session against new dc
        Session *newSession = m_api->createSession(dc);
        connect(newSession, SIGNAL(sessionReady(DC*)), this, SLOT(onUploadGetFileSessionCreated()));
        f->setSession(newSession);
        m_downloadsMap.insert(newSession->sessionId(), f);
        newSession->connectToServer();
    } else {
        Q_EMIT error(id, errorCode, errorText);
    }
}


// not direct Responses
void Telegram::onAuthSentCode(qint64 id, bool phoneRegistered, const QString &phoneCodeHash, qint32 sendCallTimeout, bool /* unused isPassword*/) {
    m_phoneCodeHash = phoneCodeHash;
    Q_EMIT authSendCodeAnswer(id, phoneRegistered, sendCallTimeout);
}

void Telegram::onContactsBlocked(qint64 id, const QList<ContactBlocked> &blocked, const QList<User> &users) {
    Q_EMIT contactsGetBlockedAnswer(id, blocked.size(), blocked, users);
}

void Telegram::onMessagesSentMessage(qint64 id, qint32 msgId, qint32 date, qint32 pts, qint32 seq) {
    QList<ContactsLink> links;
    Q_EMIT messagesSendMessageAnswer(id, msgId, date, pts, seq, links);
}

void Telegram::onMessagesSendMediaStatedMessage(qint64 id, Message message, QList<Chat> chats, QList<User> users, qint32 pts, qint32 seq) {
    QList<ContactsLink> links;
    onMessagesSendMediaStatedMessageLink(id, message, chats, users, links, pts, seq);
}

void Telegram::onMessagesSendMediaStatedMessageLink(qint64 id, Message message, QList<Chat> chats, QList<User> users, QList<ContactsLink> links, qint32 pts, qint32 seq) {
    //depending on responsed media, emit one signal or another
    switch (message.media().classType()) {
    case MessageMedia::typeMessageMediaPhoto:
        Q_EMIT messagesSendPhotoAnswer(id, message, chats, users, links, pts, seq);
        break;
    case MessageMedia::typeMessageMediaVideo:
        Q_EMIT messagesSendVideoAnswer(id, message, chats, users, links, pts, seq);
        break;
    case MessageMedia::typeMessageMediaGeo:
        Q_EMIT messagesSendGeoPointAnswer(id, message, chats, users, links, pts, seq);
        break;
    case MessageMedia::typeMessageMediaContact:
        Q_EMIT messagesSendContactAnswer(id, message, chats, users, links, pts, seq);
        break;
    case MessageMedia::typeMessageMediaDocument:
        Q_EMIT messagesSendDocumentAnswer(id, message, chats, users, links, pts, seq);
        break;
    case MessageMedia::typeMessageMediaAudio:
        Q_EMIT messagesSendAudioAnswer(id, message, chats, users, links, pts, seq);
        break;
    default:
        Q_EMIT messagesSendMediaAnswer(id, message, chats, users, links, pts, seq);
    }
}

void Telegram::onMessagesGetMessagesMessages(qint64 id, const QList<Message> &messages, const QList<Chat> &chats, const QList<User> &users) {
    Q_EMIT messagesGetMessagesAnswer(id, messages.size(), messages, chats, users);
}

void Telegram::onMessagesDialogs(qint64 id, const QList<Dialog> &dialogs, const QList<Message> &messages, const QList<Chat> &chats, const QList<User> &users) {
    Q_EMIT messagesGetDialogsAnswer(id, dialogs.size(), dialogs, messages, chats, users);
}

void Telegram::onMessagesGetHistoryMessages(qint64 id, const QList<Message> &messages, const QList<Chat> &chats, const QList<User> &users) {
    Q_EMIT messagesGetHistoryAnswer(id, messages.size(), messages, chats, users);
}

void Telegram::onMessagesSearchMessages(qint64 id, const QList<Message> &messages, const QList<Chat> &chats, const QList<User> &users) {
    Q_EMIT messagesSearchAnswer(id, messages.size(), messages, chats, users);
}

void Telegram::onMessagesForwardMsgStatedMessage(qint64 id, Message message, QList<Chat> chats, QList<User> users, qint32 pts, qint32 seq) {
    QList<ContactsLink> links;
    Q_EMIT messagesForwardMessageAnswer(id, message, chats, users, links, pts, seq);
}

void Telegram::onMessagesForwardMsgsStatedMessages(qint64 id, QList<Message> messages, QList<Chat> chats, QList<User> users, qint32 pts, qint32 seq) {
    QList<ContactsLink> links;
    Q_EMIT messagesForwardMessagesAnswer(id, messages, chats, users, links, pts, seq);
}

void Telegram::onMessagesSendBroadcastStatedMessages(qint64 id, QList<Message> messages, QList<Chat> chats, QList<User> users, qint32 pts, qint32 seq) {
    QList<ContactsLink> links;
    Q_EMIT messagesSendBroadcastAnswer(id, messages, chats, users, links, pts, seq);
}

void Telegram::onMessagesEditChatTitleStatedMessage(qint64 id, Message message, QList<Chat> chats, QList<User> users, qint32 pts, qint32 seq) {
    QList<ContactsLink> links;
    Q_EMIT messagesEditChatTitleAnswer(id, message, chats, users, links, pts, seq);
}

void Telegram::onMessagesEditChatPhotoStatedMessage(qint64 id, Message message, QList<Chat> chats, QList<User> users, qint32 pts, qint32 seq) {
    QList<ContactsLink> links;
    Q_EMIT messagesEditChatPhotoAnswer(id, message, chats, users, links, pts, seq);
}

void Telegram::onMessagesAddChatUserStatedMessage(qint64 id, Message message, QList<Chat> chats, QList<User> users, qint32 pts, qint32 seq) {
    QList<ContactsLink> links;
    Q_EMIT messagesAddChatUserAnswer(id, message, chats, users, links, pts, seq);
}

void Telegram::onMessagesDeleteChatUserStatedMessage(qint64 id, Message message, QList<Chat> chats, QList<User> users, qint32 pts, qint32 seq) {
    QList<ContactsLink> links;
    Q_EMIT messagesDeleteChatUserAnswer(id, message, chats, users, links, pts, seq);
}

void Telegram::onMessagesCreateChatStatedMessage(qint64 id, Message message, QList<Chat> chats, QList<User> users, qint32 pts, qint32 seq) {
    QList<ContactsLink> links;
    Q_EMIT messagesCreateChatAnswer(id, message, chats, users, links, pts, seq);
}

void Telegram::onUpdatesDifference(qint64 id, QList<Message> messages, QList<EncryptedMessage> newEncryptedMessages, QList<Update> otherUpdates, QList<Chat> &chats, QList<User> &users, UpdatesState state) {
    Q_EMIT updatesGetDifferenceAnswer(id, messages, newEncryptedMessages, otherUpdates, chats, users, state, false);
}

void Telegram::onUpdatesDifferenceSlice(qint64 id, QList<Message> messages, QList<EncryptedMessage> newEncryptedMessages, QList<Update> otherUpdates, QList<Chat> &chats, QList<User> &users, UpdatesState state) {
    Q_EMIT updatesGetDifferenceAnswer(id, messages, newEncryptedMessages, otherUpdates, chats, users, state, true);
}

// Requests
qint64 Telegram::authCheckPhone() {
    return m_api->authCheckPhone(Settings::getInstance()->phoneNumber());
}

qint64 Telegram::authSendCode() {
    return m_api->authSendCode(Settings::getInstance()->phoneNumber(), 0, LIBQTELEGRAM_APP_ID, LIBQTELEGRAM_APP_HASH, LANG_CODE);
}

qint64 Telegram::authSendCall() {
    return m_api->authSendCall(Settings::getInstance()->phoneNumber(), m_phoneCodeHash);
}

qint64 Telegram::authSignIn(const QString &code) {
    return m_api->authSignIn(Settings::getInstance()->phoneNumber(), m_phoneCodeHash, code);
}

qint64 Telegram::authSignUp(const QString &code, const QString &firstName, const QString &lastName) {
    return m_api->authSignUp(Settings::getInstance()->phoneNumber(), m_phoneCodeHash, code, firstName, lastName);
}

qint64 Telegram::authLogOut() {
    return m_api->authLogOut();
}

qint64 Telegram::authSendInvites(const QStringList &phoneNumbers, const QString &inviteText) {
    return m_api->authSendInvites(phoneNumbers, inviteText);
}

qint64 Telegram::authResetAuthorizations() {
    return m_api->authResetAuthorizations();
}

qint64 Telegram::accountRegisterDevice(const QString &token,  bool appSandbox) {
    return m_api->accountRegisterDevice(2, token, Utils::getDeviceModel(), Utils::getSystemVersion(), Utils::getAppVersion(), appSandbox, Settings::getInstance()->langCode());
}

qint64 Telegram::accountUnregisterDevice(const QString &token) {
    return m_api->accountUnregisterDevice(2, token);
}

qint64 Telegram::accountUpdateNotifySettings(const InputNotifyPeer &peer, const InputPeerNotifySettings &settings) {
    return m_api->accountUpdateNotifySettings(peer, settings);
}

qint64 Telegram::accountGetNotifySettings(const InputNotifyPeer &peer) {
    return m_api->accountGetNotifySettings(peer);
}

qint64 Telegram::accountResetNotifySettings() {
    return m_api->accountResetNotifySettings();
}

qint64 Telegram::accountUpdateProfile(const QString &firstName, const QString &lastName) {
    return m_api->accountUpdateProfile(firstName, lastName);
}

qint64 Telegram::accountUpdateStatus(bool offline) {
    return m_api->accountUpdateStatus(offline);
}

qint64 Telegram::accountGetWallPapers() {
    return m_api->accountGetWallPapers();
}

qint64 Telegram::photosUploadProfilePhoto(const QByteArray &bytes, const QString &fileName, const QString &caption, const InputGeoPoint &geoPoint, const InputPhotoCrop &crop) {
    FileOperation *op = new FileOperation(FileOperation::uploadProfilePhoto);
    op->setCaption(caption);
    op->setGeoPoint(geoPoint);
    op->setCrop(crop);
    return uploadSendFile(*op, fileName, bytes);
}

qint64 Telegram::photosUpdateProfilePhoto(qint64 photoId, qint64 accessHash, const InputPhotoCrop &crop) {
    InputPhoto inputPhoto(InputPhoto::typeInputPhoto);
    inputPhoto.setId(photoId);
    inputPhoto.setAccessHash(accessHash);
    return m_api->photosUpdateProfilePhoto(inputPhoto, crop);
}

qint64 Telegram::usersGetUsers(const QList<InputUser> &users) {
    return m_api->usersGetUsers(users);
}

qint64 Telegram::usersGetFullUser(const InputUser &user) {
    return m_api->usersGetFullUser(user);
}

qint64 Telegram::photosGetUserPhotos(const InputUser &user, qint32 offset, qint32 maxId, qint32 limit) {
    return m_api->photosGetUserPhotos(user, offset, maxId, limit);
}

qint64 Telegram::contactsGetStatuses() {
    return m_api->contactsGetStatuses();
}

bool lessThan(const Contact &c1, const Contact &c2) {
    return c1.userId() < c2.userId();
}

qint64 Telegram::contactsGetContacts() {
    //If there already is a full contact list on the client, an md5-hash of a comma-separated list of contact IDs
    //in ascending order may be passed in this 'hash' parameter. If the contact set was not changed,
    //contactsContactsNotModified() will be returned from Api, so the cached client list is returned with the
    //signal that they are the same contacts as previous request
    QString hash;
    if (!m_cachedContacts.isEmpty()) {
        qSort(m_cachedContacts.begin(), m_cachedContacts.end(), lessThan); //lessThan method must be outside any class or be static
        QString hashBase;
        if (m_cachedContacts.size() > 0) {
            hashBase.append(QString::number(m_cachedContacts.at(0).userId()));
        }
        for (qint32 i = 1; i < m_cachedContacts.size(); i++) {
            hashBase.append(",");
            hashBase.append(QString::number(m_cachedContacts.at(i).userId()));
        }
        QCryptographicHash md5Generator(QCryptographicHash::Md5);
        md5Generator.addData(hashBase.toStdString().c_str());
        hash = md5Generator.result().toHex();
    }
    return m_api->contactsGetContacts(hash);
}

qint64 Telegram::contactsImportContacts(const QList<InputContact> &contacts, bool replace) {
    return m_api->contactsImportContacts(contacts, replace);
}

qint64 Telegram::contactsDeleteContact(const InputUser &user) {
    return m_api->contactsDeleteContact(user);
}

qint64 Telegram::contactsDeleteContacts(const QList<InputUser> &users) {
    return m_api->contactsDeleteContacts(users);
}

qint64 Telegram::contactsBlock(const InputUser &user) {
    return m_api->contactsBlock(user);
}

qint64 Telegram::contactsUnblock(const InputUser &user) {
    return m_api->contactsUnblock(user);
}

qint64 Telegram::contactsGetBlocked(qint32 offset, qint32 limit) {
    return m_api->contactsGetBlocked(offset, limit);
}

qint64 Telegram::messagesSendMessage(const InputPeer &peer, const QString &message) {
    qint64 randomId;
    Utils::randomBytes(&randomId, 8);
    return m_api->messagesSendMessage(peer, message, randomId);
}

qint64 Telegram::messagesSendPhoto(const InputPeer &peer, const QByteArray &bytes, const QString &fileName) {
    InputMedia inputMedia(InputMedia::typeInputMediaUploadedPhoto);
    FileOperation *op = new FileOperation(FileOperation::sendMedia);
    op->setInputPeer(peer);
    op->setInputMedia(inputMedia);
    return uploadSendFile(*op, fileName, bytes);
}

qint64 Telegram::messagesSendPhoto(const InputPeer &peer, const QString &filePath) {
    InputMedia inputMedia(InputMedia::typeInputMediaUploadedPhoto);
    FileOperation *op = new FileOperation(FileOperation::sendMedia);
    op->setInputPeer(peer);
    op->setInputMedia(inputMedia);
    return uploadSendFile(*op, filePath);
}

qint64 Telegram::messagesSendGeoPoint(const InputPeer &peer, const InputGeoPoint &inputGeoPoint) {
    InputMedia inputMedia(InputMedia::typeInputMediaGeoPoint);
    inputMedia.setGeoPoint(inputGeoPoint);
    qint64 randomId;
    Utils::randomBytes(&randomId, 8);
    return m_api->messagesSendMedia(peer, inputMedia, randomId);
}

qint64 Telegram::messagesSendContact(const InputPeer &peer, const QString &phoneNumber, const QString &firstName, const QString &lastName) {
    InputMedia inputMedia(InputMedia::typeInputMediaContact);
    inputMedia.setPhoneNumber(phoneNumber);
    inputMedia.setFirstName(firstName);
    inputMedia.setLastName(lastName);
    qint64 randomId;
    Utils::randomBytes(&randomId, 8);
    return m_api->messagesSendMedia(peer, inputMedia, randomId);
}

qint64 Telegram::messagesSendVideo(const InputPeer &peer, const QByteArray &bytes, const QString &fileName, qint32 duration, qint32 width, qint32 height, const QString &mimeType, const QByteArray &thumbnailBytes, const QString &thumbnailName) {
    InputMedia inputMedia(InputMedia::typeInputMediaUploadedVideo);
    inputMedia.setDuration(duration);
    inputMedia.setW(width);
    inputMedia.setH(height);
    inputMedia.setMimeType(mimeType);
    if (!thumbnailBytes.isEmpty()) {
        inputMedia.setClassType(InputMedia::typeInputMediaUploadedThumbVideo);
    }
    FileOperation *op = new FileOperation(FileOperation::sendMedia);
    op->setInputPeer(peer);
    op->setInputMedia(inputMedia);
    return uploadSendFile(*op, fileName, bytes, thumbnailBytes, thumbnailName);
}

qint64 Telegram::messagesSendVideo(const InputPeer &peer, const QString &filePath, qint32 duration, qint32 width, qint32 height, const QString &thumbnailFilePath) {
    InputMedia inputMedia(InputMedia::typeInputMediaUploadedVideo);
    inputMedia.setDuration(duration);
    inputMedia.setW(width);
    inputMedia.setH(height);
    inputMedia.setMimeType(QMimeDatabase().mimeTypeForFile(QFileInfo(filePath)).name());
    if (thumbnailFilePath.length() > 0) {
        inputMedia.setClassType(InputMedia::typeInputMediaUploadedThumbVideo);
    }
    FileOperation *op = new FileOperation(FileOperation::sendMedia);
    op->setInputPeer(peer);
    op->setInputMedia(inputMedia);
    return uploadSendFile(*op, filePath, thumbnailFilePath);
}

qint64 Telegram::messagesSendAudio(const InputPeer &peer, const QByteArray &bytes, const QString &fileName, qint32 duration, const QString &mimeType) {
    InputMedia inputMedia(InputMedia::typeInputMediaUploadedAudio);
    inputMedia.setDuration(duration);
    inputMedia.setMimeType(mimeType);
    FileOperation *op = new FileOperation(FileOperation::sendMedia);
    op->setInputPeer(peer);
    op->setInputMedia(inputMedia);
    return uploadSendFile(*op, fileName, bytes);
}

qint64 Telegram::messagesSendAudio(const InputPeer &peer, const QString &filePath, qint32 duration) {
    InputMedia inputMedia(InputMedia::typeInputMediaUploadedAudio);
    inputMedia.setDuration(duration);
    inputMedia.setMimeType(QMimeDatabase().mimeTypeForFile(QFileInfo(filePath)).name());
    FileOperation *op = new FileOperation(FileOperation::sendMedia);
    op->setInputPeer(peer);
    op->setInputMedia(inputMedia);
    return uploadSendFile(*op, filePath);
}

qint64 Telegram::messagesSendDocument(const InputPeer &peer, const QByteArray &bytes, const QString &fileName, const QString &mimeType, const QByteArray &thumbnailBytes, const QString &thumbnailName) {
    InputMedia inputMedia(InputMedia::typeInputMediaUploadedDocument);
    inputMedia.setFileName(fileName);
    inputMedia.setMimeType(mimeType);
    if (!thumbnailBytes.isEmpty()) {
        inputMedia.setClassType(InputMedia::typeInputMediaUploadedThumbDocument);
    }
    FileOperation *op = new FileOperation(FileOperation::sendMedia);
    op->setInputPeer(peer);
    op->setInputMedia(inputMedia);
    return uploadSendFile(*op, fileName, bytes, thumbnailBytes, thumbnailName);
}

qint64 Telegram::messagesSendDocument(const InputPeer &peer, const QString &filePath, const QString &thumbnailFilePath) {
    InputMedia inputMedia(InputMedia::typeInputMediaUploadedDocument);
    inputMedia.setMimeType(QMimeDatabase().mimeTypeForFile(QFileInfo(filePath)).name());
    if (thumbnailFilePath.length() > 0) {
        inputMedia.setClassType(InputMedia::typeInputMediaUploadedThumbDocument);
    }
    FileOperation *op = new FileOperation(FileOperation::sendMedia);
    op->setInputPeer(peer);
    op->setInputMedia(inputMedia);
    return uploadSendFile(*op, filePath, thumbnailFilePath);
}

qint64 Telegram::messagesForwardPhoto(const InputPeer &peer, qint64 photoId, qint64 accessHash) {
    InputPhoto inputPhoto(InputPhoto::typeInputPhoto);
    inputPhoto.setId(photoId);
    inputPhoto.setAccessHash(accessHash);
    InputMedia inputMedia(InputMedia::typeInputMediaPhoto);
    inputMedia.setPhotoId(inputPhoto);
    qint64 randomId;
    Utils::randomBytes(&randomId, 8);
    return m_api->messagesSendMedia(peer, inputMedia, randomId);
}

qint64 Telegram::messagesForwardVideo(const InputPeer &peer, qint64 videoId, qint64 accessHash) {
    InputVideo inputVideo(InputVideo::typeInputVideo);
    inputVideo.setId(videoId);
    inputVideo.setAccessHash(accessHash);
    InputMedia inputMedia(InputMedia::typeInputMediaVideo);
    inputMedia.setVideoId(inputVideo);
    qint64 randomId;
    Utils::randomBytes(&randomId, 8);
    return m_api->messagesSendMedia(peer, inputMedia, randomId);
}

qint64 Telegram::messagesForwardAudio(const InputPeer &peer, qint64 audioId, qint64 accessHash) {
    InputAudio inputAudio(InputAudio::typeInputAudio);
    inputAudio.setId(audioId);
    inputAudio.setAccessHash(accessHash);
    InputMedia inputMedia(InputMedia::typeInputMediaAudio);
    inputMedia.setAudioId(inputAudio);
    qint64 randomId;
    Utils::randomBytes(&randomId, 8);
    return m_api->messagesSendMedia(peer, inputMedia, randomId);
}

qint64 Telegram::messagesForwardDocument(const InputPeer &peer, qint64 documentId, qint64 accessHash) {
    InputDocument inputDocument(InputDocument::typeInputDocument);
    inputDocument.setId(documentId);
    inputDocument.setAccessHash(accessHash);
    InputMedia inputMedia(InputMedia::typeInputMediaDocument);
    inputMedia.setDocumentId(inputDocument);
    qint64 randomId;
    Utils::randomBytes(&randomId, 8);
    return m_api->messagesSendMedia(peer, inputMedia, randomId);
}

qint64 Telegram::messagesSetTyping(const InputPeer &peer, bool typing) {
    return m_api->messagesSetTyping(peer, typing);
}

qint64 Telegram::messagesGetMessages(const QList<qint32> &msgIds) {
    return m_api->messagesGetMessages(msgIds);
}

qint64 Telegram::messagesGetDialogs(qint32 offset, qint32 maxId, qint32 limit) {
    return m_api->messagesGetDialogs(offset, maxId, limit);
}

qint64 Telegram::messagesGetHistory(const InputPeer &peer, qint32 offset, qint32 maxId, qint32 limit) {
    return m_api->messagesGetHistory(peer, offset, maxId, limit);
}

qint64 Telegram::messagesSearch(const InputPeer &peer, const QString &query, MessagesFilter filter, qint32 minDate, qint32 maxDate, qint32 offset, qint32 maxId, qint32 limit) {
    return m_api->messagesSearch(peer, query, filter, minDate, maxDate, offset, maxId, limit);
}

qint64 Telegram::messagesReadHistory(const InputPeer &peer, qint32 maxId, qint32 offset) {
    return m_api->messagesReadHistory(peer, maxId, offset);
}

qint64 Telegram::messagesDeleteHistory(const InputPeer &peer, qint32 offset) {
    return m_api->messagesDeleteHistory(peer, offset);
}

qint64 Telegram::messagesDeleteMessages(const QList<qint32> &msgIds) {
    return m_api->messagesDeleteMessages(msgIds);
}

qint64 Telegram::messagesRestoreMessages(const QList<qint32> &msgIds) {
    return m_api->messagesRestoreMessages(msgIds);
}

qint64 Telegram::messagesReceivedMessages(qint32 maxId) {
    return m_api->messagesReceivedMessages(maxId);
}

qint64 Telegram::messagesForwardMessage(const InputPeer &peer, qint32 msgId) {
    qint64 randomId;
    Utils::randomBytes(&randomId, 8);
    return m_api->messagesForwardMessage(peer, msgId, randomId);
}

qint64 Telegram::messagesForwardMessages(const InputPeer &peer, const QList<qint32> &msgIds) {
    return m_api->messagesForwardMessages(peer, msgIds);
}

qint64 Telegram::messagesSendBroadcast(const QList<InputUser> &users, const QString &message, const InputMedia &media) {
    return m_api->messagesSendBroadcast(users, message, media);
}

qint64 Telegram::messagesGetChats(const QList<qint32> &chatIds) {
    return m_api->messagesGetChats(chatIds);
}

qint64 Telegram::messagesGetFullChat(qint32 chatId) {
    return m_api->messagesGetFullChat(chatId);
}

qint64 Telegram::messagesEditChatTitle(qint32 chatId, const QString &title) {
    return m_api->messagesEditChatTitle(chatId, title);
}

qint64 Telegram::messagesEditChatPhoto(qint32 chatId, const QByteArray &bytes, const QString &fileName, const InputPhotoCrop &crop) {
    InputChatPhoto inputChatPhoto(InputChatPhoto::typeInputChatUploadedPhoto);
    inputChatPhoto.setCrop(crop);
    FileOperation *op = new FileOperation(FileOperation::editChatPhoto);
    op->setChatId(chatId);
    op->setInputChatPhoto(inputChatPhoto);
    return uploadSendFile(*op, fileName, bytes);
}

qint64 Telegram::messagesEditChatPhoto(qint32 chatId, qint64 photoId, qint64 accessHash, const InputPhotoCrop &crop) {
    InputChatPhoto inputChatPhoto(InputChatPhoto::typeInputChatPhoto);
    InputPhoto inputPhoto(InputPhoto::typeInputPhoto);
    inputPhoto.setId(photoId);
    inputPhoto.setAccessHash(accessHash);
    inputChatPhoto.setId(inputPhoto);
    inputChatPhoto.setCrop(crop);
    return m_api->messagesEditChatPhoto(chatId, inputChatPhoto);
}

qint64 Telegram::messagesAddChatUser(qint32 chatId, const InputUser &user, qint32 fwdLimit) {
    return m_api->messagesAddChatUser(chatId, user, fwdLimit);
}

qint64 Telegram::messagesDeleteChatUser(qint32 chatId, const InputUser &user) {
    return m_api->messagesDeleteChatUser(chatId, user);
}

qint64 Telegram::messagesCreateChat(const QList<InputUser> &users, const QString &chatTopic) {
    return m_api->messagesCreateChat(users, chatTopic);
}

qint64 Telegram::updatesGetState() {
    return m_api->updatesGetState();
}

qint64 Telegram::updatesGetDifference(qint32 pts, qint32 date, qint32 qts) {
    return m_api->updatesGetDifference(pts, date, qts);
}


// private methods
qint64 Telegram::uploadSendFile(const FileOperation &op, const QString &fileName, const QByteArray &bytes, const QByteArray &thumbnailBytes, const QString &thumbnailName) {
    // create dedicated session to working dc for the main file
    DC *dc = m_dcProvider.getWorkingDc();
    Session *session = m_api->createSession(dc);
    UploadFile *f = new UploadFile(session, UploadFile::file, bytes, this);
    f->setName(fileName);
    m_filesMap.insert(f->id(), f);
    m_fileOperationsMap.insert(f->id(), &op);
    // if exists thumbnail, create its file object and assign id to main file thumbnail_id (and viceversa)
    // first send all thumbnail before sending main file
    if (!thumbnailBytes.isEmpty()) {
        // use same session for thumbnail and main file. In case a thumbnail exists, send it first
        UploadFile *thumb = new UploadFile(session, UploadFile::thumbnail, thumbnailBytes, this);
        f->setRelatedFileId(thumb->id());
        thumb->setRelatedFileId(f->id());
        thumbnailName.length() > 0 ? thumb->setName(thumbnailName) : thumb->setName(fileName);
        // insert the relationship fileId <-> thumb file
        m_filesMap.insert(thumb->id(), thumb);
        // insert linked to session id the thumb file to be uploaded next
        m_filesMap.insert(session->sessionId(), thumb);
    } else {
        // link to session id the file to be uploaded after session creation (sessionReady signal)
        m_filesMap.insert(session->sessionId(), f);
    }

    connect(session, SIGNAL(sessionReady(DC*)), this, SLOT(onUploadSendFileSessionCreated()));
    session->connectToServer();
    return f->id();
}

qint64 Telegram::uploadSendFile(const FileOperation &op, const QString &filePath, const QString &thumbnailPath) {
    // create dedicated session to working dc for the main file
    DC *dc = m_dcProvider.getWorkingDc();
    Session *session = m_api->createSession(dc);
    UploadFile *f = new UploadFile(session, UploadFile::file, filePath, this);
    m_filesMap.insert(f->id(), f);
    m_fileOperationsMap.insert(f->id(), &op);
    // if exists thumbnail, create its file object and assign id to main file thumbnail_id (and viceversa)
    // first send all thumbnail before sending main file
    if (thumbnailPath.length() > 0) {
        // use same session for thumbnail and main file. In case a thumbnail exists, send it first
        UploadFile *thumb = new UploadFile(session, UploadFile::thumbnail, thumbnailPath, this);
        f->setRelatedFileId(thumb->id());
        thumb->setRelatedFileId(f->id());
        // insert the relationship fileId <-> thumb file
        m_filesMap.insert(thumb->id(), thumb);
        // insert linked to session id the thumb file to be uploaded next
        m_filesMap.insert(session->sessionId(), thumb);
    } else {
        // link to session id the file to be uploaded after session creation (sessionReady signal)
        m_filesMap.insert(session->sessionId(), f);
    }

    connect(session, SIGNAL(sessionReady(DC*)), this, SLOT(onUploadSendFileSessionCreated()));
    session->connectToServer();
    return f->id();
}

void Telegram::onUploadSendFileSessionCreated() {
    Session *session = qobject_cast<Session*>(sender());
    UploadFile *f = m_filesMap.take(session->sessionId());
    uploadSendFileParts(*f);
}

qint64 Telegram::uploadSendFileParts(UploadFile &file) {
    // depending on size use one or other api method to upload every part
    qint32 partId = 0;
    if (file.length() > 10 * 1024 * 1024) {
        while (file.hasMoreParts()) {
            QByteArray bytes = file.nextPart();
            m_api->uploadSaveBigFilePart(file.session(), file.id(), partId++, file.nParts(), bytes);
        }
    } else {
        while (file.hasMoreParts()) {
            QByteArray bytes = file.nextPart();
            m_api->uploadSaveFilePart(file.session(), file.id(), partId++, bytes);
        }
    }
    // return file id instead of every part msg id
    return file.id();
}

void Telegram::onUploadSaveFilePartResult(qint64, qint64 fileId, bool) {
    UploadFile *f = m_filesMap.value(fileId);
    Q_ASSERT(f);
    f->increaseUploadedParts();
    qDebug() << "uploaded" << f->uploadedParts() << "parts out of" << f->nParts();
    if (f->uploadedParts() == f->nParts()) {
        // if finished a thumbnail, send the main file. Send media metadata if finished is the main file
        qDebug() << "file upload finished for fileId" << fileId;
        if (f->fileType() == UploadFile::thumbnail) {
            qint64 mainFileId = f->relatedFileId();
            UploadFile *main = m_filesMap.value(mainFileId);
            Q_ASSERT(main);
            uploadSendFileParts(*main);
        } else {

            // compose InputFile to be added to metadata
            InputFile inputFile(InputFile::typeInputFile);
            if (f->length() > 10 * 1024 * 1024) {
                inputFile.setClassType(InputFile::typeInputFileBig);
            } else {
                inputFile.setMd5Checksum(QString(f->md5().toHex()));
            }
            inputFile.setId(fileId);
            inputFile.setParts(f->nParts());
            inputFile.setName(f->name());

            // Read operation and execute internal api operation depending on file operation type
            const FileOperation *op = m_fileOperationsMap.take(f->id());
            switch (op->opType()) {
                case FileOperation::sendMedia: {
                    InputMedia metadata = op->inputMedia();
                    metadata.setFile(inputFile);
                    // once filled input main file, check if there is a thumbnail and fill its data
                    if (f->relatedFileId()) {
                        UploadFile *thumb = m_filesMap.take(f->relatedFileId());
                        InputFile inputThumb(InputFile::typeInputFile);
                        if (thumb->length() > 10 * 1024 * 1024) {
                            inputThumb.setClassType(InputFile::typeInputFileBig);
                        } else {
                            inputThumb.setMd5Checksum(QString(thumb->md5().toHex()));
                        }
                        inputThumb.setId(thumb->id());
                        inputThumb.setParts(thumb->nParts());
                        inputThumb.setName(thumb->name());
                        metadata.setThumb(inputThumb);
                        delete thumb;
                    }
                    // retrieve peer related to file, generate random id and send media metadata
                    InputPeer peer = op->peer();
                    qint64 randomId;
                    Utils::randomBytes(&randomId, 8);
                    m_api->messagesSendMedia(peer, metadata, randomId);
                    break;
                }
                case FileOperation::editChatPhoto: {
                    InputChatPhoto metadata = op->inputChatPhoto();
                    metadata.setFile(inputFile);
                    qint32 chatId = op->chatId();
                    m_api->messagesEditChatPhoto(chatId, metadata);
                    break;
                }
                case FileOperation::uploadProfilePhoto: {
                    QString caption = op->caption();
                    InputGeoPoint geoPoint = op->geoPoint();
                    InputPhotoCrop crop = op->crop();
                    m_api->photosUploadProfilePhoto(inputFile, caption, geoPoint, crop);
                    break;
                }
            }
            // delete main file from map after sending
            m_filesMap.remove(fileId);
            // delete operation and main file pointers
            delete f;
            delete op;

        }
    }
}

qint64 Telegram::uploadGetFile(const InputFileLocation &location, qint32 fileSize, qint32 dcNum) {
    // change of dc if received a dcNum
    DC *dc;
    dcNum ? dc = m_dcProvider.getDc(dcNum) : dc = m_dcProvider.getWorkingDc();
    Session *session = m_api->createSession(dc);
    DownloadFile *f = new DownloadFile(session, location, fileSize, this);
    m_downloadsMap.insert(session->sessionId(), f);
    connect(session, SIGNAL(sessionReady(DC*)), this, SLOT(onUploadGetFileSessionCreated()));
    session->connectToServer();
    // return file id as request id for apps
    return f->id();
}

void Telegram::onUploadGetFileSessionCreated() {
    Session *session = qobject_cast<Session *>(sender());
    DownloadFile *f = m_downloadsMap.take(session->sessionId());
    qint64 msgId = m_api->uploadGetFile(session, f->fileLocation());
    m_downloadsMap.insert(msgId, f);
}

void Telegram::onUploadGetFileAnswer(qint64 msgId, StorageFileType type, qint32 mtime, QByteArray bytes) {
    DownloadFile *f = m_downloadsMap.take(msgId);
    Q_ASSERT(f);
    if (type.classType() == StorageFileType::typeStorageFilePartial) {
        // if 'managedDownloads' flag is set, store all temporal received data and return a unique event
        // when file downloading is finished. Emit an event for every part otherwise.
        if (Settings::getInstance()->managedDownloads()) {
            // partial bytes appended to DownloadFile object
            f->appendBytes(bytes);
            if (f->bytes().length() < f->expectedSize()) {
                qint64 newMsgId = m_api->uploadGetFile(f->session(), f->fileLocation(), f->offset(), f->partLength());
                m_downloadsMap.insert(newMsgId, f);
            } else {
                Q_EMIT uploadGetFileAnswer(f->id(), type, mtime, f->bytes(), 0, f->length(), f->expectedSize()); //emit signal of finished
                delete f;
            }
        } else {
            qint32 downloaded = f->partId() * f->partLength() + bytes.length();
            Q_EMIT uploadGetFileAnswer(f->id(), type, mtime, bytes, f->partId(), downloaded, f->expectedSize());
            f->incrementPartId();
            f->setOffset(f->offset() + bytes.length());
            if (f->offset() >= f->expectedSize()) {
                delete f;
            } else {
                qint64 newMsgId = m_api->uploadGetFile(f->session(), f->fileLocation(), f->offset(), f->partLength());
                m_downloadsMap.insert(newMsgId, f);
            }
        }

    } else {
        Q_EMIT uploadGetFileAnswer(f->id(), type, mtime, bytes, 0, bytes.length(), f->expectedSize());
        delete f;
    }

}
