#include "data.h"
#include "dbmanager.h"
#include <QtSql/QSqlQuery>
#include <QtSql/QSqlError>
#include <QVariant>
#include "rawapiclient.h"

Data::Data(QObject *parent) :
    QObject(parent) {
}

Data *Data::instance() {
    static Data* self = new Data;
    return self;
}


void Data::onMessagesGetDialogsAnswer(qint64, qint32 totalCount, const QList<Dialog> &dialogs, const QList<Message> &messages, const QList<Chat> &chats, const QList<User> &users) {

    // local storage
    DbManager::instance()->beginTransaction();

    Q_FOREACH (User user, users) {
        if (!insertOrUpdateUser(user)) {
            DbManager::instance()->rollbackTransaction();
            qCWarning(TG_PLUGIN_LOGIC) << "Error inserting dialogs mentioned users in local database";
            return;
        }
    }

    Q_FOREACH (Chat chat, chats) {
        if (!insertOrUpdateChat(chat)) {
            DbManager::instance()->rollbackTransaction();
            qCWarning(TG_PLUGIN_LOGIC) << "Error inserting dialogs mentioned chats in local database";
            return;
        }
    }

    Q_FOREACH (Message message, messages) {
        if (!insertOrUpdateMessage(message)) {
            DbManager::instance()->rollbackTransaction();
            qCWarning(TG_PLUGIN_LOGIC) << "Error inserting dialogs top messages in local database";
            return;
        }
    }

    Q_FOREACH (Dialog dialog, dialogs) {
        if (!insertOrUpdateDialog(dialog)) {
            DbManager::instance()->rollbackTransaction();
            qCWarning(TG_PLUGIN_LOGIC) << "Error inserting dialogs main info in local database";
            return;
        }
    }

    DbManager::instance()->finishTransaction();

    // after local storage, compose the events to send to models
    QList<ContactItem *> contacts;
    QList<DialogItem *> dlgs;

    // create maps for mentioned users and chats to be easily referenced when composing dialog items.
    // Afterwards, use this map creation to compose contacts list also
    QMap<qint32, const User*> mentionedUsers;
    for (qint32 i=0; i<users.size(); i++) {
        const User &u = users.at(i);
        mentionedUsers.insert(u.id(), &u);
        ContactItem *c = new ContactItem(this);
        c->setId(u.id());
        c->setFirstName(u.firstName());
        c->setLastName(u.lastName());
        c->setPhone(u.phone());
        c->setStatus(u.status());
        contacts.append(c);
    }

    Q_EMIT contactsAdded(contacts);

    QMap<qint32, const Chat*> mentionedChats;
    for (qint32 i=0; i<chats.size(); i++) {
        const Chat &c = chats.at(i);
        mentionedChats.insert(c.id(), &c);
    }

    for (qint32 i = 0; i < dialogs.size(); ++i) {
        DialogItem *item = new DialogItem(this);
        Dialog thisdialog = dialogs.at(i);
        Message thismessage = messages.at(i);
        Peer peer = thisdialog.peer();
        FileLocation dialogPhoto(FileLocation::typeFileLocationUnavailable);
        if (peer.classType() == Peer::typePeerChat) {
            const Chat *c = mentionedChats.value(peer.chatId());
            item->setIsChat(true);
            item->setId(c->id());
            if (c->classType() != Chat::typeChatEmpty) {
                item->setName(c->title());
                if (c->classType() != Chat::typeChatForbidden) {
                    dialogPhoto = c->photo().photoSmall();
                }
            }
        } else if (peer.classType() == Peer::typePeerUser) {
            const User *user = mentionedUsers.value(peer.userId());
            item->setIsChat(false);
            item->setId(peer.userId());
            // store phone as default value in dialog. When rendered the dialog, user name will be searched
            // in contacts model, but if not user found, at least its phone is shown
            item->setName(user->phone());
            dialogPhoto = user->photo().photoSmall();
        }
        /*XXX the way of retrieving contacts and dialogs photos will be performed later. By now commented
        // retrieve user or chat photo asynchronously
        if (dialogPhoto.classType() != FileLocation::typeFileLocationUnavailable) {
            InputFileLocation f(InputFileLocation::typeInputFileLocation);
            f.setLocalId(dialogPhoto.localId());
            f.setSecret(dialogPhoto.secret());
            f.setVolumeId(dialogPhoto.volumeId());
            qint64 fileId = m_telegram->uploadGetFile(f, 0, dialogPhoto.dcId());
            m_photoFiles.insert(fileId, item->id());
        }*/
        item->setUnreadCount(thisdialog.unreadCount());

        // append the top message
        MessageItem* messageItem = new MessageItem(thismessage);
        item->appendMessage(messageItem);

        dlgs.append(item);
    }

    Q_EMIT dialogsAdded(dlgs, totalCount);
}

QList<DialogItem *> Data::getDialogs() {

    QList<DialogItem *> result;

    QSqlQuery query(DbManager::instance()->database());

    query.prepare("SELECT id, isChat, unreadCount FROM dialogs");
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "error retrieving dialogs list from database";
        return result;
    }

    while (query.next()) {
        DialogItem *d = new DialogItem(this);
        d->setId(query.value("id").toInt());
        d->setIsChat(query.value("isChat").toBool());
        d->setUnreadCount(query.value("unreadCount").toInt());

        if (d->isChat()) {
            DialogItem *chat = getChat(d->id());
            d->setName(chat->name());
            d->setMembers(chat->members());
            d->setThumbnail(chat->thumbnail());
            chat->deleteLater();
        } else {
            ContactItem *user = getUser(d->id());
            d->setName(user->firstName() + " " + user->lastName());
            d->setThumbnail(user->thumbnail());
            d->setUserOnline(user->status()->classType() == UserStatus::typeUserStatusOnline);
            d->setLastSeenOnline(user->status()->wasOnline());
        }

        MessageItem *message = topMessage(d->id(), d->isChat());
        MessagesModel &mm = d->getMessagesModel();
        mm.append(message);
        result.append(d);
    }
    return result;
}

DialogItem *Data::getDialog(qint32 dialogId) {

    DialogItem *d;

    QSqlQuery query(DbManager::instance()->database());

    query.prepare("SELECT isChat, unreadCount FROM dialogs WHERE id");
    query.bindValue(":id", dialogId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "queried dialog with id" << dialogId << "has not been found";
        return d;
    }
    query.next();
    d = new DialogItem(this);
    d->setId(dialogId);
    d->setIsChat(query.value("isChat").toBool());
    d->setUnreadCount(query.value("unreadCount").toInt());
    return d;
}

MessageItem *Data::topMessage(qint32 dialogId, bool isChat) {

    MessageItem *m;

    QSqlQuery query(DbManager::instance()->database());

    if (isChat) {
        query.prepare("SELECT id, fromId, toId, toIsChat, out, unread, date, text, mediaType "
                      "FROM messages "
                      "WHERE toId=:dialogId AND toIsChat=1 "
                      "ORDER BY datetime(date) DESC LIMIT 1");
    } else {
        query.prepare("SELECT id, fromId, toId, toIsChat, out, unread, date, text, mediaType "
                      "FROM messages "
                      "WHERE toIsChat=0 AND (toId=:dialogId OR fromId=:dialogId) "
                      "ORDER BY datetime(date) DESC LIMIT 1");
    }
    query.bindValue(":dialogId", dialogId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "queried top message in dialog with id" << dialogId << "has not been found:" << query.lastError() << query.lastQuery();;
        return m;
    }
    query.next();
    m = new MessageItem(this);
    m->setId(query.value("id").toInt());
    m->setFromId(query.value("fromId").toInt());
    Peer p(Peer::typePeerUser);
    if (query.value("toIsChat").toBool()) {
        p.setChatId(query.value("toId").toInt());
    } else {
        p.setUserId(query.value("toId").toInt());
    }
    m->setToId(p);
    m->setOut(query.value("out").toBool());
    m->setUnread(query.value("unread").toBool());
    m->setDate(query.value("date").toInt());
    m->setMessage(query.value("text").toString());
    MessageMedia media((MessageMedia::MessageMediaType)query.value("mediaType").toInt());
    m->setMedia(media);
    return m;
}

bool Data::insertOrUpdateDialog(const Dialog &dialog) {
    if (!insertDialog(dialog)) {
        return updateDialog(dialog);
    }
    return true;
}

bool Data::insertDialog(const Dialog &dialog) {

    bool isChat = dialog.peer().classType() == Peer::typePeerChat;
    qint32 dialogId = isChat ? dialog.peer().chatId() : dialog.peer().userId();

    QSqlQuery query(DbManager::instance()->database());

    query.prepare("INSERT INTO dialogs (id, isChat, unreadCount) VALUES (:id, :isChat, :unreadCount)");
    query.bindValue(":id", dialogId);
    query.bindValue(":isChat", isChat);
    query.bindValue(":unreadCount", dialog.unreadCount());
    return query.exec();
}

bool Data::updateDialog(const Dialog &dialog) {

    bool isChat = dialog.peer().classType() == Peer::typePeerChat;
    qint32 dialogId = isChat ? dialog.peer().chatId() : dialog.peer().userId();

    QSqlQuery query(DbManager::instance()->database());

    query.prepare("UPDATE dialogs SET isChat=:isChat, unreadCount=:unreadCount WHERE id=:id");
    query.bindValue(":id", dialogId);
    query.bindValue(":isChat", isChat);
    query.bindValue(":unreadCount", dialog.unreadCount());
    bool ok = query.exec();
    if (!ok) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
    }
    return ok;
}

bool Data::deleteDialogs() {

    DbManager::instance()->beginTransaction();

    QSqlQuery query(DbManager::instance()->database());
    query.prepare("DELETE FROM dialogs");

    if (query.exec()) {
        if (deleteMessages()) {
            DbManager::instance()->finishTransaction();
            return true;
        } else {
            DbManager::instance()->rollbackTransaction();
        }
    } else {
        DbManager::instance()->rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
    }
    return false;
}

bool Data::deleteDialog(qint32 dialogId) {

    DbManager::instance()->beginTransaction();

    QSqlQuery query(DbManager::instance()->database());
    query.prepare("DELETE FROM dialogs WHERE id=:id");
    query.bindValue(":id", dialogId);

    if (query.exec()) {
        if (deleteMessages(dialogId)) {
            DbManager::instance()->finishTransaction();
        } else {
            DbManager::instance()->rollbackTransaction();
        }
    } else {
        DbManager::instance()->rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
    }
}

// MESSAGEs operations
MessageItem *Data::getMessage(qint32 messageId) {
    MessageItem *m;

    QSqlQuery query(DbManager::instance()->database());

    query.prepare("SELECT fromId, toId, toIsChat, out, unread, date, text, mediaType "
                  "FROM messages WHERE id=:id");
    query.bindValue(":id", messageId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "queried message with id" << messageId << "has not been found";
        return m;
    }
    query.next();
    m = new MessageItem(this);
    m->setId(messageId);
    m->setFromId(query.value("fromId").toInt());
    Peer p(Peer::typePeerUser);
    if (query.value("toIsChat").toBool()) {
        p.setChatId(query.value("toId").toInt());
    } else {
        p.setUserId(query.value("toId").toInt());
    }
    m->setToId(p);
    m->setOut(query.value("out").toBool());
    m->setUnread(query.value("unread").toBool());
    m->setDate(query.value("date").toInt());
    m->setMessage(query.value("text").toString());
    MessageMedia media((MessageMedia::MessageMediaType)query.value("mediaType").toInt());
    m->setMedia(media);
    return m;
}

bool Data::insertOrUpdateMessage(const Message &message) {
    if (!insertMessage(message)) {
        return updateMessage(message);
    }
    return true;
}

bool Data::updateMessage(const Message &message) {
    bool toIsChat = message.toId().classType() == Peer::typePeerChat;
    qint32 toId = toIsChat ? message.toId().chatId() : message.toId().userId();
    qint32 mediaType = message.media().classType();

    QSqlQuery query(DbManager::instance()->database());

    // try to append message to database
    query.prepare("UPDATE messages "
                  "SET fromId=:fromId, toId=:toId, toIsChat=:toIsChat, out=:out, unread=:unread, date=:date, text=:text, mediaType=:mediaType "
                  "WHERE id=:id");
    query.bindValue(":id", message.id());
    query.bindValue(":fromId", message.fromId());
    query.bindValue(":toId", toId);
    query.bindValue(":toIsChat", toIsChat);
    query.bindValue(":out", message.out());
    query.bindValue(":unread", message.unread());
    query.bindValue(":date", message.date());
    query.bindValue(":text", message.message());
    query.bindValue(":mediaType", mediaType);
    bool ok = query.exec();
    if (!ok) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
    }
    return ok;
}

bool Data::insertMessage(const Message &message) {

    bool toIsChat = message.toId().classType() == Peer::typePeerChat;
    qint32 toId = toIsChat ? message.toId().chatId() : message.toId().userId();
    qint32 mediaType = message.media().classType();

    QSqlQuery query(DbManager::instance()->database());

    // try to append message to database
    query.prepare("INSERT INTO messages (id, fromId, toId, toIsChat, out, unread, date, text, mediaType) "
                  "VALUES (:id, :fromId, :toId, :toIsChat, :out, :unread, :date, :text, :mediaType)");
    query.bindValue(":id", message.id());
    query.bindValue(":fromId", message.fromId());
    query.bindValue(":toId", toId);
    query.bindValue(":toIsChat", toIsChat);
    query.bindValue(":out", message.out());
    query.bindValue(":unread", message.unread());
    query.bindValue(":date", message.date());
    query.bindValue(":text", message.message());
    query.bindValue(":mediaType", mediaType);
    return query.exec();
}

bool Data::deleteMessages() {

    QSqlQuery query(DbManager::instance()->database());

    query.prepare("DELETE FROM messages");
    bool ok = query.exec();
    if (!ok) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
    }
    return ok;
}

bool Data::deleteMessages(qint32 dialogId) {

    QSqlQuery query(DbManager::instance()->database());

    query.prepare("SELECT isChat FROM dialogs WHERE id=:dialogId");
    query.bindValue(":dialogId", dialogId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
        return false;
    }

    query.next();
    // if is chat
    if (query.value(0).toBool()) {
        query.prepare("DELETE FROM messages WHERE toId=:dialogId AND toIsChat=true");
    } else {
        query.prepare("DELETE FROM messages WHERE toIsChat=false AND (toId=:dialogId OR fromId=:dialogId)");
    }
    query.bindValue(":dialogId", dialogId);
    bool ok = query.exec();
    if (!ok) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
    }
    return ok;;
}

bool Data::markMessageAsRead(qint32 messageId) {

    QSqlQuery query(DbManager::instance()->database());

    query.prepare("UPDATE messages SET unread=false WHERE id=:id");
    query.bindValue(":id", messageId);
    bool ok = query.exec();
    if (!ok) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
    }
    return ok;
}


// USERs operations
ContactItem *Data::getUser(qint32 id) {

    ContactItem *c;

    QSqlQuery query(DbManager::instance()->database());

    query.prepare("SELECT firstName, lastName, phone, thumbnail, lastSeenOnline, online "
                  "FROM users WHERE id=:id");
    query.bindValue(":id", id);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "queried user with id" << id << "has not been found";
        return c;
    }
    query.next();
    c = new ContactItem(this);
    c->setId(id);
    c->setFirstName(query.value("firstName").toString());
    c->setLastName(query.value("lastName").toString());
    c->setPhone(query.value("phone").toString());
    c->setThumbnail(query.value("thumbnail").toString());

    UserStatus us(UserStatus::typeUserStatusOffline);
    if (query.value("online").toBool()) {
        us.setClassType(UserStatus::typeUserStatusOnline);
    }
    us.setWasOnline(query.value("lastSeenOnline").toInt());
    c->setStatus(us);
    return c;
}

QList<ContactItem *> Data::getUsers() {
    QList<ContactItem *> result;

    QSqlQuery query(DbManager::instance()->database());

    query.prepare("SELECT id, firstName, lastName, phone, thumbnail, lastSeenOnline, online FROM users");
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "an error has happened when requesting complete contacts list";
        return result;
    }
    while (query.next()) {
        ContactItem *c = new ContactItem(this);
        c->setId(query.value("id").toInt());
        c->setFirstName(query.value("firstName").toString());
        c->setLastName(query.value("lastName").toString());
        c->setPhone(query.value("phone").toString());
        c->setThumbnail(query.value("thumbnail").toString());
        UserStatus us(UserStatus::typeUserStatusOffline);
        if (query.value("online").toBool()) {
            us.setClassType(UserStatus::typeUserStatusOnline);
        }
        us.setWasOnline(query.value("lastSeenOnline").toInt());
        c->setStatus(us);
        result.append(c);
    }
    return result;
}

bool Data::insertOrUpdateUser(const User &user) {
    if (!insertUser(user)) {
        return updateUser(user);
    }
    return true;
}

bool Data::insertUser(const User &user) {

    bool online = user.status().classType() == UserStatus::typeUserStatusOnline;

    QSqlQuery query(DbManager::instance()->database());

    // try to append user to database
    query.prepare("INSERT INTO users (id, firstName, lastName, phone, lastSeenOnline, online) "
                  "VALUES (:id, :firstName, :lastName, :phone, :lastSeenOnline, :online)");
    query.bindValue(":id", user.id());
    query.bindValue(":firstName", user.firstName());
    query.bindValue(":lastName", user.lastName());
    query.bindValue(":phone", user.phone());
    query.bindValue(":lastSeenOnline", user.status().wasOnline());
    query.bindValue(":online", online);
    return query.exec();
}

bool Data::updateUser(const User &user) {

    bool online = user.status().classType() == UserStatus::typeUserStatusOnline;

    QSqlQuery query(DbManager::instance()->database());

    // try to update existant
    query.prepare("UPDATE users SET firstName=:firstName, lastName=:lastName, phone=:phone, lastSeenOnline=:lastSeenOnline, online=:online "
                  "WHERE id=:id");
    query.bindValue(":firstName", user.firstName());
    query.bindValue(":lastName", user.lastName());
    query.bindValue(":phone", user.phone());
    query.bindValue(":lastSeenOnline", user.status().wasOnline());
    query.bindValue(":online", online);
    query.bindValue(":id", user.id());
    bool ok = query.exec();
    if (!ok) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
    }
    return ok;
}

bool Data::updateUserThumbnail(qint32 userId, const QString &thumbnail) {

    QSqlQuery query(DbManager::instance()->database());

    // try to update the user
    query.prepare("UPDATE users SET thumbnail=:thumbnail WHERE users.id=:id");
    query.bindValue(":thumbnail", thumbnail);
    query.bindValue("id", userId);
    bool ok = query.exec();
    if (!ok) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
    }
    return ok;
}

bool Data::deleteUsers() {

    QSqlQuery query(DbManager::instance()->database());

    query.prepare("DELETE FROM users");
    bool ok = query.exec();
    if (!ok) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
    }
    return ok;
}

// CHATs operations
DialogItem *Data::getChat(qint32 id) {
    DialogItem *d;

    QSqlQuery query(DbManager::instance()->database());

    query.prepare("SELECT title, participantsCount, thumbnail FROM chats WHERE id=:id");
    query.bindValue(":id", id);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "queried chat with id" << id << "has not been found";
        return d;
    }
    query.next();
    d = new DialogItem(this);
    d->setId(id);
    d->setName(query.value("title").toString());
    d->setMembers(query.value("participantsCount").toInt());
    d->setThumbnail(query.value("thumbnail").toString());
    return d;
}

bool Data::insertOrUpdateChat(const Chat &chat) {
    if (!insertChat(chat)) {
        return updateChat(chat);
    }
    return true;
}

bool Data::insertChat(const Chat &chat) {

    QSqlQuery query(DbManager::instance()->database());

    // try to append user to database
    query.prepare("INSERT INTO chats (id, title, participantsCount) VALUES (:id, :title, :participantsCount)");
    query.bindValue(":id", chat.id());
    query.bindValue(":title", chat.title());
    query.bindValue(":participantsCount", chat.participantsCount());
    return query.exec();
}

bool Data::updateChat(const Chat &chat) {

    QSqlQuery query(DbManager::instance()->database());

    // try to append user to database
    query.prepare("UPDATE chats SET title=:title, participantsCount=:participantsCount "
                  "WHERE id=:id");
    query.bindValue(":id", chat.id());
    query.bindValue(":title", chat.title());
    query.bindValue(":participantsCount", chat.participantsCount());
    bool ok = query.exec();
    if (!ok) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
    }
    return ok;
}

bool Data::deleteChat(qint32 chatId) {

    QSqlQuery query(DbManager::instance()->database());

    query.prepare("DELETE FROM chats WHERE id=:id");
    query.bindValue(":id", chatId);
    bool ok = query.exec();
    if (!ok) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
    }
    return ok;
}
