#include "dbmanager.h"
#include <QDir>
#include <QFile>
#include <QtSql/QSqlQuery>
#include <QtSql/QSqlError>
#include <QtSql/QSqlDriver>
#include "config.h"


DbManager::DbManager(QObject *parent) :
    QObject(parent) {
}

DbManager::~DbManager() {
    if (m_database.isOpen()) {
        m_database.close();
    }
}

DbManager *DbManager::instance() {
    static DbManager *self = new DbManager;
    return self;
}

bool DbManager::initializeDatabase() {

    QDir dir(Config::instance()->databasePath());
    if (!dir.exists()) {
        qCWarning(TG_PLUGIN_LOGIC) << "path where create the database does not exists";
        return false;
    }

    m_databasePath = dir.absoluteFilePath(DATABASE_NAME);
    m_database = QSqlDatabase::addDatabase("QSQLITE");
    m_database.setDatabaseName(m_databasePath);

    if (!createOrUpdateDatabase()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Failed to create or update the database";
        return false;
    }
    return true;
}

QSqlDatabase DbManager::database() const {
    return m_database;
}

bool DbManager::beginTransaction()
{
    return m_database.transaction();
}

bool DbManager::finishTransaction()
{
    return m_database.commit();
}

bool DbManager::rollbackTransaction()
{
    return m_database.rollback();
}

bool DbManager::createOrUpdateDatabase() {

    bool create = !QFile(m_databasePath).exists();

    QStringList statements;
    if (create) {
        statements = parseSchemaFile(SCHEMA_FILE);

        if (statements.isEmpty()) {
            return true;
        }

        if (!m_database.open()) {
            return false;
        }

        QSqlQuery query(m_database);

        beginTransaction();

        Q_FOREACH(const QString &statement, statements) {
            if (!query.exec(statement)) {
                qCCritical(TG_PLUGIN_LOGIC) << "Failed to create or update database. SQL Statements:" << query.lastQuery() << "Error:" << query.lastError();
                rollbackTransaction();
                return false;
            }
        }

        finishTransaction();
    }

    if (!m_database.open()) {
        return false;
    }

    return true;

}

QStringList DbManager::parseSchemaFile(const QString &fileName)
{
    QFile schema(fileName);
    if (!schema.open(QFile::ReadOnly)) {
        qCCritical(TG_PLUGIN_LOGIC) << "Failed to open " << fileName;
        return QStringList();
    }

    bool parsingBlock = false;
    QString statement;
    QStringList statements;

    // FIXME: this parser is very basic, it needs to be improved in the future
    //        it does a lot of assumptions based on the structure of the schema.sql file

    QTextStream stream(&schema);
    while (!stream.atEnd()) {
        QString line = stream.readLine();
        bool statementEnded = false;

        statement += line;

        // check if we are parsing a trigger command
        if (line.trimmed().startsWith("CREATE TRIGGER", Qt::CaseInsensitive)) {
            parsingBlock = true;
        } else if (parsingBlock) {
            if (line.contains("END;")) {
                parsingBlock = false;
                statementEnded = true;
            }
        } else if (statement.contains(";")) {
            statementEnded = true;
        }

        statement += "\n";

        if (statementEnded) {
            statements.append(statement);
            statement.clear();
        }
    }

    return statements;
}


