#include "contactsmodel.h"
#include "data.h"

ContactsModel::ContactsModel(QObject *parent) :
    QAbstractListModel(parent) {

    connect(Data::instance(), SIGNAL(contactsAdded(QList<ContactItem*>)), this, SLOT(onContactsAdded(QList<ContactItem*>)));
}


QHash<int, QByteArray> ContactsModel::roleNames() const {
    QHash<int, QByteArray> roles;
    roles[IdRole] = "id";
    roles[FirstNameRole] = "firstName";
    roles[LastNameRole] = "lastName";
    roles[PhoneRole] = "phone";
    roles[ThumbnailRole] = "thumbnail";
    roles[LastSeenOnlineRole] = "lastSeenOnline";
    roles[OnlineRole] = "online";
    return roles;
}

int ContactsModel::rowCount(const QModelIndex &parent) const {

    if (parent.isValid()) {
        return 0;
    }

    return m_contacts.count();
}

QVariant ContactsModel::data(const QModelIndex &index, int role) const {

    if (!index.isValid() || index.row() < 0 || index.row() >= m_contacts.count()) {
        return QVariant();
    }

    ContactItem *c = m_contacts.at(index.row());
    QVariant result;

    switch (role) {
    case IdRole:
        result = c->id();
        break;
    case FirstNameRole:
        result = c->firstName();
        break;
    case LastNameRole:
        result = c->lastName();
        break;
    case PhoneRole:
        result = c->phone();
        break;
    case ThumbnailRole:
        result = c->thumbnail();
        break;
    case LastSeenOnlineRole:
        result = c->lastSeenOnline();
        break;
    case OnlineRole:
        result = c->online();
        break;
    }

    return result;
}

void ContactsModel::append(ContactItem *contact) {
    // if this contact exists, update, otherwise append at the end
    qint32 pos = m_contactIdsMap.value(contact->id());
    if (pos) {
        m_contacts.replace(pos, contact);
    } else {
        beginInsertRows(QModelIndex(), m_contacts.count(), m_contacts.count());
        m_contactIdsMap.insert(contact->id(), m_contacts.count());
        m_contacts.append(contact);
        endInsertRows();
    }
}

void ContactsModel::clear() {
    beginResetModel();
    m_contacts.clear();
    m_contactIdsMap.clear();
    endResetModel();
}

ContactItem *ContactsModel::contact(qint32 id) {
    Q_ASSERT(id);
    Q_ASSERT(m_contacts.count() > 0);
    qint32 pos = m_contactIdsMap.value(id);
    return m_contacts.at(pos);
}

void ContactsModel::updateThumbnail(qint32 id, const QString &thumbnailPath) {
    qint32 pos = m_contactIdsMap.value(id);
    if ((pos >= 0) && (pos < m_contacts.count())) {
        m_contacts.at(pos)->setThumbnail(thumbnailPath);
        const QVector<qint32> changedRoles( QVector<qint32>() << ThumbnailRole);
        Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    }
}


// SLOTS
void ContactsModel::onContactsAdded(const QList<ContactItem *> &contacts) {

    // fill a list with the new contacts not existing yet in model
    QList<ContactItem *> contactsToAppend;

    // replace contact or add to contacts to append list
    Q_FOREACH (ContactItem *c, contacts) {
        qint32 pos = m_contactIdsMap.value(c->id(), -1);
        if (pos >= 0) {
            if (m_contacts.at(pos)) {
                m_contacts.at(pos)->deleteLater();
            }
            m_contacts.replace(pos, c);
        } else {
            contactsToAppend.append(c);
        }

        // temporally dump of object to see if valid
        c->dump();
    }

    Q_FOREACH (ContactItem *c, contactsToAppend) {
        append(c);
    }
}
