#include "dialogsmodel.h"

#include "telegramclient.h"
#include "types/tlmessagemedia.h"
#include "data.h"
#include <QFile>
#include <QDir>

#include "modelsmanager.h"


DialogsModel::DialogsModel(QObject *parent) :
    QAbstractListModel(parent),
    m_canFetchMore(true) {
    // register message item type to be used as QVariant in qml code
    qRegisterMetaType<MessageItem>("MessageItem");

    connect(Data::instance(), SIGNAL(dialogsAdded(QList<DialogItem*>,qint32)), this, SLOT(onDialogsAdded(QList<DialogItem*>,qint32)));
}

DialogsModel::~DialogsModel() {
    Q_FOREACH (DialogItem* item, m_dialogs) {
        if (item) {
            delete item;
            item = 0;
        }
    }
}

QHash<int, QByteArray> DialogsModel::roleNames() const {
    QHash<int, QByteArray> roles;
    roles[IdRole] = "id";
    roles[IsChatRole] = "isChat";
    roles[NameRole] = "name";
    roles[UnreadCountRole] = "unreadCount";
    roles[ThumbnailRole] = "thumbnail";
    roles[TypingRole] = "typing";
    roles[WhoisTypingRole] = "whoisTyping";
    roles[MembersRole] = "membersRole";
    roles[MembersOnlineRole] = "membersOnline";
    roles[UserOnlineRole] = "userOnline";
    roles[LastSeenOnlineRole] = "lastSeenOnline";
    roles[TopMessageRole] = "topMessage";
    return roles;
}

qint32 DialogsModel::rowCount(const QModelIndex &parent) const {

    if (parent.isValid()) {
        return 0;
    }

    return m_dialogs.size();
}

QVariant DialogsModel::get(int row) {
    if ((row < 0) || (row >= m_dialogs.size())) {
        return QVariant();
    }
    return QVariant::fromValue(m_dialogs.at(row));
}

QVariant DialogsModel::getMessagesModel(int row) {
    if ((row < 0) || (row >= m_dialogs.size())) {
        return QVariant();
    }
    DialogItem *tlDialog = m_dialogs.at(row);
    return QVariant::fromValue(&tlDialog->getMessagesModel());
}

QVariant DialogsModel::getMessagesModelFromId(int dialogId) {
    qint32 pos = m_dlgIdsMap.value(dialogId);
    return getMessagesModel(pos);
}

QVariant DialogsModel::data(const QModelIndex &index, int role) const {

    if (!index.isValid() || index.row() < 0 || index.row() >= m_dialogs.count()) {
        return QVariant();
    }

    DialogItem *d = m_dialogs.at(index.row());

    QVariant result;

    switch (role) {
    case IdRole:
        result = d->id();
        break;
    case IsChatRole:
        result = d->isChat();
        break;
    case NameRole:
        result = d->name();
        break;
    case UnreadCountRole:
        result = d->unreadCount();
        break;
    case ThumbnailRole:
        result = d->thumbnail();
        break;
    case TypingRole:
        result = d->typing();
        break;
    case WhoisTypingRole:
        result = d->whoisTyping();
        break;
    case MembersRole:
        result = d->members();
        break;
    case MembersOnlineRole:
        result = d->membersOnline();
        break;
    case UserOnlineRole:
        result = d->userOnline();
        break;
    case LastSeenOnlineRole:
        result = d->lastSeenOnline();
        break;
    case TopMessageRole:
        result = QVariant::fromValue(d->topMessage());
        break;
    }
    return result;
}

Qt::ItemFlags DialogsModel::flags(const QModelIndex &index) const {
    if (!index.isValid()) {
        return Qt::ItemIsEnabled;
    }
    return QAbstractItemModel::flags(index) /*| Qt::ItemIsEditable*/;
}


void DialogsModel::append(DialogItem *dialog) {
    beginInsertRows(QModelIndex(), m_dialogs.count(), m_dialogs.count());
    m_dlgIdsMap.insert(dialog->id(), m_dialogs.count());
    m_dialogs.append(dialog);
    endInsertRows();
}

void DialogsModel::clear() {
    beginResetModel();
    m_dialogs.clear();
    m_dlgIdsMap.clear();
    m_topMessageIdsMap.clear();
    endResetModel();
}


// SLOTS
void DialogsModel::onDialogsAdded(const QList<DialogItem *> &dialogs, qint32 totalCount) {

    QList<DialogItem *> dialogsToAppend;

    // edit existing ones
    Q_FOREACH (DialogItem *d, dialogs) {
        qint32 pos = m_dlgIdsMap.value(d->id(), -1);
        if (pos >= 0) {
            if (m_dialogs.at(pos)) {
                m_dialogs.at(pos)->deleteLater();
            }
            m_dialogs.replace(pos, d);
        } else {
            dialogsToAppend.append(d);
        }

        // temporally dump of the object to see if valid
        d->dump();
    }

    Q_FOREACH (DialogItem *d, dialogsToAppend) {
        append(d);
    }

    m_canFetchMore = totalCount > m_dialogs.count();
}


void DialogsModel::updateThumbnail(qint32 dialogId, QString filePath) {
    // update thumbnail path for affected dialogsItem
    qint32 pos = m_dlgIdsMap.value(dialogId);
    DialogItem *d = m_dialogs.value(pos);
    d->setThumbnail(filePath);
    //emit signal thumbnail changed
    const QVector<qint32> changedRoles( QVector<qint32>() << ThumbnailRole );
    Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
}

void DialogsModel::updateReceivedTopMessage(qint32 dialogId, qint32 messageId, const QString &text, qint32 date, qint32 fromId) {
    qint32 pos = m_dlgIdsMap.value(dialogId);
    DialogItem *d = m_dialogs.value(pos);
    d->setUnreadCount(d->unreadCount() + 1);
    if (!d->topMessage()) {
        d->createTopMessage();
    }
    d->topMessage()->setId(messageId);
    d->topMessage()->setMessage(text);
    d->topMessage()->setDate(date);
    d->topMessage()->setFromId(fromId);
    d->topMessage()->setUnread(true);
    d->topMessage()->setOut(false);
    const QVector<qint32> changedRoles( QVector<qint32>() << UnreadCountRole << TopMessageRole);
    Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    // update map messageId -> dialogsItemId
    m_topMessageIdsMap.insert(messageId, dialogId);
    // void set typing for this dialog
    d->setIsTyping(0);
}

void DialogsModel::updateSendingTopMessage(qint32 dialogId, qint32 requestId, const QString &text) {
    qint32 pos = m_dlgIdsMap.value(dialogId);
    DialogItem *d = m_dialogs.value(pos);
    if (!d->topMessage()) {
        d->createTopMessage();
    };
    d->topMessage()->setId(requestId);
    d->topMessage()->setMessage(text);
    d->topMessage()->setOut(true);
    d->topMessage()->setUnread(true);
    d->topMessage()->setFromId(Settings::getInstance()->ourId());
    const QVector<qint32> changedRoles( QVector<qint32>() << TopMessageRole);
    Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    // insert relationship between sending request id and dialogItemId
    m_topMessageIdsMap.insert(requestId, dialogId);
}

void DialogsModel::updateSentTopMessage(qint32 requestId, qint32 messageId, qint32 date) {
    // get requestId related dialogItemId
    qint32 dialogId = m_topMessageIdsMap.take(requestId);
    qint32 pos = m_dlgIdsMap.value(dialogId);
    DialogItem *d = m_dialogs.value(pos);
    if (!d->topMessage()) {
        d->createTopMessage();
    };
    d->topMessage()->setId(messageId);
    d->topMessage()->setDate(date);
    d->topMessage()->setOut(true);
    d->topMessage()->setUnread(true);
    d->topMessage()->setFromId(Settings::getInstance()->ourId());
    const QVector<qint32> changedRoles( QVector<qint32>() << TopMessageRole);
    Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    // update map messageId -> dialogsItemId
    m_topMessageIdsMap.insert(messageId, dialogId);
}

//XXX this method must be rewritten to check for the message amongst all messages in same dialog
void DialogsModel::markAsRead(qint32 messageId) {
    qint32 dialogId = m_topMessageIdsMap.take(messageId);
    if (dialogId) {
        qint32 pos = m_dlgIdsMap.value(dialogId);
        DialogItem *d = m_dialogs.value(pos);
        d->topMessage()->setUnread(false);
        QVector<qint32> changedRoles( QVector<qint32>() << TopMessageRole);
        // if top message is not sent by me, set unreadcount to 0
        if (d->topMessage()->fromId() != Settings::getInstance()->ourId()) {
            d->setUnreadCount(0);
            changedRoles.append(UnreadCountRole);
        }
        Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    }
}

void DialogsModel::setIsTyping(qint32 dialogId, qint32 userId) {
    qint32 pos = m_dlgIdsMap.value(dialogId);
    DialogItem *d = m_dialogs.value(pos);
    d->setIsTyping(userId);
    const QVector<qint32> changedRoles( QVector<qint32>() << TypingRole << WhoisTypingRole);
    Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    // set signal slot connection to aware when timeout is fired
    connect(d, SIGNAL(typingFinished()), this, SLOT(onTypingFinished()));
}

void DialogsModel::onTypingFinished() {
    DialogItem *d = qobject_cast<DialogItem *>(sender());
    qint32 pos = m_dlgIdsMap.value(d->id());
    const QVector<qint32> changedRoles( QVector<qint32>() << TypingRole << WhoisTypingRole);
    Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    // if typing is empty, disconnect signal-slot
    disconnect(d, SIGNAL(typingFinished()), this, SLOT(onTypingFinished()));
}
